"""
IR mocks
"""
from collections import OrderedDict
from copy import deepcopy

from cdd.ast_utils import NoneStr
from cdd.defaults_utils import remove_defaults_from_intermediate_repr
from cdd.pure_utils import deindent, paren_wrap_code
from cdd.tests.mocks.classes import (
    class_torch_nn_l1loss_docstring_str,
    class_torch_nn_one_cycle_lr_docstring_str,
    tensorboard_doc_str_no_args_str,
)
from cdd.tests.mocks.docstrings import (
    docstring_google_pytorch_lbfgs_str,
    docstring_google_tf_adadelta_str,
    docstring_google_tf_adam_str,
    docstring_google_tf_lambda_callback_str,
    docstring_google_tf_squared_hinge_no_args_doc_str,
    docstring_header_str,
)
from cdd.tests.utils_for_tests import remove_args_from_docstring

class_google_tf_tensorboard_ir = {
    "doc": deindent(tensorboard_doc_str_no_args_str, sep="  ", level=1).rstrip("\n"),
    "name": None,
    "params": OrderedDict(
        (
            (
                "log_dir",
                {
                    "default": "logs",
                    "doc": "the path of the directory where to save the log "
                    "files to be parsed by TensorBoard.",
                    "typ": "str",
                },
            ),
            (
                "histogram_freq",
                {
                    "default": 0,
                    "doc": "frequency (in epochs) at which to compute activation "
                    "and weight histograms for the layers of the model. "
                    "If set to 0, histograms won't be computed. "
                    "Validation data (or split) must be specified for "
                    "histogram visualizations.",
                    "typ": "int",
                },
            ),
            (
                "write_graph",
                {
                    "default": True,
                    "doc": "whether to visualize the graph in TensorBoard. The "
                    "log file can become quite large when write_graph is "
                    "set to True.",
                    "typ": "bool",
                },
            ),
            (
                "write_images",
                {
                    "default": False,
                    "doc": "whether to write model weights to visualize as image "
                    "in TensorBoard.",
                    "typ": "bool",
                },
            ),
            (
                "update_freq",
                {
                    "default": "epoch",
                    "doc": "`'batch'` or `'epoch'` or integer. When using "
                    "`'batch'`, writes the losses and metrics to "
                    "TensorBoard after each batch. The same applies for "
                    "`'epoch'`. If using an integer, let's say `1000`, "
                    "the callback will write the metrics and losses to "
                    "TensorBoard every 1000 batches. Note that writing "
                    "too frequently to TensorBoard can slow down your "
                    "training.",
                    "typ": "str",
                },
            ),
            (
                "profile_batch",
                {
                    "default": 2,
                    "doc": "Profile the batch(es) to sample compute "
                    "characteristics. profile_batch must be a "
                    "non-negative integer or a tuple of integers. A pair "
                    "of positive integers signify a range of batches to "
                    "profile. By default, it will profile the second "
                    "batch. Set profile_batch=0 to disable profiling.",
                    "typ": "int",
                },
            ),
            (
                "embeddings_freq",
                {
                    "default": 0,
                    "doc": "frequency (in epochs) at which embedding layers will "
                    "be visualized. If set to 0, embeddings won't be "
                    "visualized.",
                    "typ": "int",
                },
            ),
            (
                "embeddings_metadata",
                {
                    "default": NoneStr,
                    "doc": "a dictionary which maps layer name to a file name in "
                    "which metadata for this embedding layer is saved. "
                    "See the [details]( "
                    "https://www.tensorflow.org/how_tos/embedding_viz/#metadata_optional) "
                    "about metadata files format. In case if the same "
                    "metadata file is used for all embedding layers, "
                    "string can be passed.",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}

class_torch_nn_l1loss_ir = {
    "doc": deindent(remove_args_from_docstring(class_torch_nn_l1loss_docstring_str), 1),
    "name": None,
    "params": OrderedDict(
        (
            (
                "size_average",
                {
                    "default": True,
                    "doc": "Deprecated (see :attr:`reduction`). By default, the "
                    "losses are averaged over each loss element in the "
                    "batch. Note that for some losses, there are multiple "
                    "elements per sample. If the field "
                    ":attr:`size_average` is set to ``False``, the losses "
                    "are instead summed for each minibatch. Ignored when "
                    "reduce is ``False``.",
                    "typ": "Optional[bool]",
                },
            ),
            (
                "reduce",
                {
                    "default": True,
                    "doc": "Deprecated (see :attr:`reduction`). By default, the "
                    "losses are averaged or summed over observations for "
                    "each minibatch depending on :attr:`size_average`. "
                    "When :attr:`reduce` is ``False``, returns a loss per "
                    "batch element instead and ignores "
                    ":attr:`size_average`.",
                    "typ": "Optional[bool]",
                },
            ),
            (
                "reduction",
                {
                    "default": "mean",
                    "doc": "Specifies the reduction to apply to the output: "
                    "``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no "
                    "reduction will be applied, ``'mean'``: the sum of "
                    "the output will be divided by the number of elements "
                    "in the output, ``'sum'``: the output will be summed. "
                    "Note: :attr:`size_average` and :attr:`reduce` are in "
                    "the process of being deprecated, and in the "
                    "meantime, specifying either of those two args will "
                    "override :attr:`reduction`.",
                    "typ": "Optional[string]",
                },
            ),
            ("__constants__", {"default": ["reduction"], "typ": "List"}),
        )
    ),
    "returns": OrderedDict((("return_type", {"typ": "None"}),)),
    "type": "static",
}

class_torch_nn_one_cycle_lr_ir = {
    "doc": deindent(
        remove_args_from_docstring(class_torch_nn_one_cycle_lr_docstring_str), 1
    ),
    "name": None,
    "params": OrderedDict(
        (
            ("optimizer", {"doc": "Wrapped optimizer.", "typ": "Optimizer"}),
            (
                "max_lr",
                {
                    "doc": "Upper learning rate boundaries in the cycle "
                    "for each parameter group.",
                    "typ": "Union[float, list]",
                },
            ),
            (
                "total_steps",
                {
                    "default": NoneStr,
                    "doc": "The total number of steps in the cycle. Note "
                    "that if a value is not provided here, then it "
                    "must be inferred by providing a value for "
                    "epochs and steps_per_epoch.",
                    "typ": "int",
                },
            ),
            (
                "epochs",
                {
                    "default": NoneStr,
                    "doc": "The number of epochs to train for. This is "
                    "used along with steps_per_epoch in order to "
                    "infer the total number of steps in the cycle "
                    "if a value for total_steps is not provided.",
                    "typ": "int",
                },
            ),
            (
                "steps_per_epoch",
                {
                    "default": NoneStr,
                    "doc": "The number of steps per epoch to train for. "
                    "This is used along with epochs in order to "
                    "infer the total number of steps in the cycle "
                    "if a value for total_steps is not provided.",
                    "typ": "int",
                },
            ),
            (
                "pct_start",
                {
                    "default": 0.3,
                    "doc": "The percentage of the cycle (in number of "
                    "steps) spent increasing the learning rate.",
                    "typ": "float",
                },
            ),
            (
                "anneal_strategy",
                {
                    "default": "cos",
                    "doc": 'Specifies the annealing strategy: "cos" for '
                    'cosine annealing, "linear" for linear '
                    "annealing.",
                    "typ": "Literal['cos', 'linear']",
                },
            ),
            (
                "cycle_momentum",
                {
                    "default": True,
                    "doc": "If ``True``, momentum is cycled inversely to "
                    "learning rate between 'base_momentum' and "
                    "'max_momentum'.",
                    "typ": "bool",
                },
            ),
            (
                "base_momentum",
                {
                    "default": 0.85,
                    "doc": "Lower momentum boundaries in the cycle for "
                    "each parameter group. Note that momentum is "
                    "cycled inversely to learning rate; at the peak "
                    "of a cycle, momentum is 'base_momentum' and "
                    "learning rate is 'max_lr'.",
                    "typ": "Union[float, list]",
                },
            ),
            (
                "max_momentum",
                {
                    "default": 0.95,
                    "doc": "Upper momentum boundaries in the cycle for "
                    "each parameter group. Functionally, it defines "
                    "the cycle amplitude (max_momentum - "
                    "base_momentum). Note that momentum is cycled "
                    "inversely to learning rate; at the start of a "
                    "cycle, momentum is 'max_momentum' and learning "
                    "rate is 'base_lr'",
                    "typ": "Union[float, list]",
                },
            ),
            (
                "div_factor",
                {
                    "default": 25.0,
                    "doc": "Determines the initial learning rate via "
                    "initial_lr = max_lr/div_factor",
                    "typ": "float",
                },
            ),
            (
                "final_div_factor",
                {
                    "default": 10000.0,
                    "doc": "Determines the minimum learning rate via "
                    "min_lr = initial_lr/final_div_factor",
                    "typ": "float",
                },
            ),
            (
                "last_epoch",
                {
                    "default": -1,
                    "doc": "The index of the last batch. This parameter is "
                    "used when resuming a training job. Since "
                    "`step()` should be invoked after each batch "
                    "instead of after each epoch, this number "
                    "represents the total number of *batches* "
                    "computed, not the total number of epochs "
                    "computed. When last_epoch=-1, the schedule is "
                    "started from the beginning.",
                    "typ": "int",
                },
            ),
            (
                "verbose",
                {
                    "default": False,
                    "doc": "If ``True``, prints a message to stdout for "
                    "each update.",
                    "typ": "bool",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}

docstring_google_tf_adadelta_ir = {
    "doc": remove_args_from_docstring(docstring_google_tf_adadelta_str),
    "name": None,
    "params": OrderedDict(
        (
            (
                "learning_rate",
                {
                    "doc": "A `Tensor`, floating point "
                    "value, or a schedule that is a "
                    "`tf.keras.optimizers.schedules.LearningRateSchedule`. "
                    "The learning rate. To match the "
                    "exact form in the original paper "
                    "use 1.0."
                },
            ),
            (
                "rho",
                {"doc": "A `Tensor` or a floating point " "value. The decay rate."},
            ),
            (
                "epsilon",
                {
                    "doc": "A `Tensor` or a floating point "
                    "value.  A constant epsilon used "
                    "to better conditioning the grad "
                    "update."
                },
            ),
            (
                "name",
                {
                    "default": "Adadelta",
                    "doc": "Optional name prefix for the "
                    "operations created when applying "
                    "gradients.  Defaults to "
                    '`"Adadelta"`.',
                    "typ": "Optional[str]",
                },
            ),
            (
                "kwargs",
                {
                    "default": NoneStr,
                    "doc": "Keyword arguments. Allowed to be "
                    'one of `"clipnorm"` or '
                    '`"clipvalue"`. `"clipnorm"` '
                    "(float) clips gradients by norm; "
                    '`"clipvalue"` (float) clips '
                    "gradients by value.",
                    "typ": "Optional[dict]",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}


docstring_google_tf_lambda_callback_ir = {
    "doc": remove_args_from_docstring(docstring_google_tf_lambda_callback_str),
    "name": None,
    "params": OrderedDict(
        (
            ("on_epoch_begin", {"doc": "called at the beginning of every " "epoch."}),
            ("on_epoch_end", {"doc": "called at the end of every epoch."}),
            ("on_batch_begin", {"doc": "called at the beginning of every " "batch."}),
            ("on_batch_end", {"doc": "called at the end of every batch."}),
            (
                "on_train_begin",
                {"doc": "called at the beginning of model " "training."},
            ),
            ("on_train_end", {"doc": "called at the end of model " "training."}),
        )
    ),
    "returns": None,
    "type": "static",
}

docstring_google_tf_adadelta_function_ir = {
    "doc": remove_args_from_docstring(docstring_google_tf_adadelta_str),
    "name": "Adadelta",
    "params": OrderedDict(
        (
            (
                "learning_rate",
                {
                    "default": 0.001,
                    "doc": "A `Tensor`, floating point "
                    "value, or a schedule that is a "
                    "`tf.keras.optimizers.schedules.LearningRateSchedule`. "
                    "The learning rate. To match the "
                    "exact form in the original paper "
                    "use 1.0.",
                    "typ": "float",
                },
            ),
            (
                "rho",
                {
                    "default": 0.95,
                    "doc": "A `Tensor` or a floating point " "value. The decay rate.",
                    "typ": "float",
                },
            ),
            (
                "epsilon",
                {
                    "default": 1e-07,
                    "doc": "A `Tensor` or a floating point "
                    "value.  A constant epsilon used "
                    "to better conditioning the grad "
                    "update.",
                    "typ": "float",
                },
            ),
            (
                "name",
                {
                    "default": "Adadelta",
                    "doc": "Optional name prefix for the "
                    "operations created when applying "
                    "gradients.",
                    "typ": "Optional[str]",
                },
            ),
            (
                "kwargs",
                {
                    "default": NoneStr,
                    "doc": "Keyword arguments. Allowed to be "
                    'one of `"clipnorm"` or '
                    '`"clipvalue"`. `"clipnorm"` '
                    "(float) clips gradients by norm; "
                    '`"clipvalue"` (float) clips '
                    "gradients by value.",
                    "typ": "Optional[dict]",
                },
            ),
            ("_HAS_AGGREGATE_GRAD", {"default": True, "typ": "bool"}),
        )
    ),
    "returns": None,
}

docstring_google_tf_adam_ir = {
    "doc": remove_args_from_docstring(docstring_google_tf_adam_str),
    "name": None,
    "params": OrderedDict(
        (
            (
                "learning_rate",
                {
                    "default": 0.001,
                    "doc": "A `Tensor`, floating point "
                    "value, or a schedule that is a "
                    "`tf.keras.optimizers.schedules.LearningRateSchedule`, "
                    "or a callable that takes no "
                    "arguments and returns the actual "
                    "value to use, The learning rate. "
                    "Defaults to 0.001.",
                    "typ": "float",
                },
            ),
            (
                "beta_1",
                {
                    "default": 0.9,
                    "doc": "A float value or a constant "
                    "float tensor, or a callable that "
                    "takes no arguments and returns "
                    "the actual value to use. The "
                    "exponential decay rate for the "
                    "1st moment estimates. Defaults "
                    "to 0.9.",
                    "typ": "float",
                },
            ),
            (
                "beta_2",
                {
                    "default": 0.999,
                    "doc": "A float value or a constant "
                    "float tensor, or a callable that "
                    "takes no arguments and returns "
                    "the actual value to use, The "
                    "exponential decay rate for the "
                    "2nd moment estimates. Defaults "
                    "to 0.999.",
                    "typ": "float",
                },
            ),
            (
                "epsilon",
                {
                    "default": 1e-07,
                    "doc": "A small constant for numerical "
                    "stability. This epsilon is "
                    '"epsilon hat" in the Kingma and '
                    "Ba paper (in the formula just "
                    "before Section 2.1), not the "
                    "epsilon in Algorithm 1 of the "
                    "paper. Defaults to 1e-7.",
                    "typ": "float",
                },
            ),
            (
                "amsgrad",
                {
                    "default": False,
                    "doc": "Boolean. Whether to apply "
                    "AMSGrad variant of this "
                    'algorithm from the paper "On the '
                    'Convergence of Adam and beyond". '
                    "Defaults to `False`.",
                    "typ": "bool",
                },
            ),
            (
                "name",
                {
                    "default": "Adam",
                    "doc": "Optional name for the operations "
                    "created when applying gradients. "
                    'Defaults to `"Adam"`.',
                    "typ": "Optional[str]",
                },
            ),
            (
                "kwargs",
                {
                    "default": NoneStr,
                    "doc": "Keyword arguments. Allowed to be "
                    'one of `"clipnorm"` or '
                    '`"clipvalue"`. `"clipnorm"` '
                    "(float) clips gradients by norm; "
                    '`"clipvalue"` (float) clips '
                    "gradients by value.",
                    "typ": "Optional[dict]",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}

docstring_google_tf_squared_hinge_ir = {
    "doc": docstring_google_tf_squared_hinge_no_args_doc_str,
    "name": None,
    "params": OrderedDict(
        (
            (
                "reduction",
                {
                    "default": "AUTO",
                    "doc": "(Optional) Type of "
                    "`tf.keras.losses.Reduction` to "
                    "apply to loss. Default value is "
                    "`AUTO`. `AUTO` indicates that "
                    "the reduction option will be "
                    "determined by the usage context. "
                    "For almost all cases this "
                    "defaults to "
                    "`SUM_OVER_BATCH_SIZE`. When used "
                    "with `tf.distribute.Strategy`, "
                    "outside of built-in training "
                    "loops such as `tf.keras` "
                    "`compile` and `fit`, using "
                    "`AUTO` or `SUM_OVER_BATCH_SIZE` "
                    "will raise an error. Please see "
                    "this custom training [tutorial]( "
                    "https://www.tensorflow.org/tutorials/distribute/custom_training) "
                    "for more details.",
                    "typ": "Optional[str]",
                },
            ),
            (
                "name",
                {
                    "default": "squared_hinge",
                    "doc": "Optional name for the op. " "Defaults to 'squared_hinge'.",
                    "typ": "Optional[str]",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}

docstring_google_pytorch_lbfgs_ir = {
    "doc": remove_args_from_docstring(docstring_google_pytorch_lbfgs_str).strip(),
    "name": None,
    "params": OrderedDict(
        (
            ("lr", {"default": 1.0, "doc": "learning rate", "typ": "float"}),
            (
                "max_iter",
                {
                    "default": 20,
                    "doc": "maximal number of iterations per optimization step",
                    "typ": "int",
                },
            ),
            (
                "max_eval",
                {
                    "default": "```max_iter * 1.25```",
                    "doc": "maximal number of function evaluations per "
                    "optimization step",
                },
            ),
            (
                "tolerance_grad",
                {
                    "default": 1e-05,
                    "doc": "termination tolerance on first order optimality",
                    "typ": "float",
                },
            ),
            (
                "tolerance_change",
                {
                    "default": 1e-09,
                    "doc": "termination tolerance on function value/parameter "
                    "changes",
                    "typ": "float",
                },
            ),
            (
                "history_size",
                {"default": 100, "doc": "update history size", "typ": "int"},
            ),
            (
                "line_search_fn",
                {
                    "default": NoneStr,
                    "doc": "either 'strong_wolfe' or None",
                    "typ": "str",
                },
            ),
        )
    ),
    "returns": None,
    "type": "static",
}

function_adder_ir = {
    "doc": "",
    "name": "add_6_5",
    "params": OrderedDict(
        (
            ("a", {"default": 6, "doc": "first param", "typ": "int"}),
            ("b", {"default": 5, "doc": "second param", "typ": "int"}),
        )
    ),
    "returns": OrderedDict(
        (
            (
                "return_type",
                {
                    "default": "```operator.add(a, b)```",
                    "doc": "Aggregated summation " "of `a` and `b`.",
                },
            ),
        )
    ),
    "type": "static",
}

method_complex_args_variety_ir = {
    "doc": "Call cliff",
    "name": "call_cliff",
    "params": OrderedDict(
        (
            ("dataset_name", {"doc": "name of dataset."}),
            ("as_numpy", {"doc": "Convert to numpy ndarrays."}),
            (
                "K",
                {
                    "doc": "backend engine, e.g., `np` or " "`tf`.",
                    "typ": "Literal['np', 'tf']",
                },
            ),
            (
                "tfds_dir",
                {
                    "default": "~/tensorflow_datasets",
                    "doc": "directory to look for models in.",
                    "typ": "str",
                },
            ),
            (
                "writer",
                {
                    "default": "```{}```".format(paren_wrap_code("stdout")),
                    "doc": "IO object to write out to",
                },
            ),
            (
                "kwargs",
                {
                    "doc": "additional keyword arguments",
                    "typ": "Optional[dict]",
                    "default": NoneStr,
                },
            ),
        )
    ),
    "returns": OrderedDict(
        (
            (
                "return_type",
                {"default": "K", "doc": "backend engine", "typ": "Literal['np', 'tf']"},
            ),
        )
    ),
    "type": "self",
}

intermediate_repr_extra_colons = {
    "name": None,
    "params": OrderedDict(
        (
            (
                "dataset_name",
                {"doc": "Example: foo", "typ": "str"},
            ),
        )
    ),
    "returns": None,
    "doc": "Some comment",
    "type": "static",
}

intermediate_repr_no_default_doc = {
    "name": None,
    "type": "static",
    "doc": docstring_header_str.rstrip("\n"),
    "params": OrderedDict(
        (
            (
                "dataset_name",
                {"default": "mnist", "doc": "name of dataset.", "typ": "str"},
            ),
            (
                "tfds_dir",
                {
                    "default": "~/tensorflow_datasets",
                    "doc": "directory to look for models in.",
                    "typ": "str",
                },
            ),
            (
                "K",
                {
                    "default": "np",
                    "doc": "backend engine, e.g., `np` or " "`tf`.",
                    "typ": "Literal['np', 'tf']",
                },
            ),
            (
                "as_numpy",
                {
                    "default": NoneStr,
                    "doc": "Convert to numpy ndarrays.",
                    "typ": "Optional[bool]",
                },
            ),
            (
                "data_loader_kwargs",
                {
                    "default": NoneStr,
                    "doc": "pass this as arguments to " "data_loader function",
                    "typ": "Optional[dict]",
                },
            ),
        )
    ),
    "returns": OrderedDict(
        (
            (
                "return_type",
                {
                    "default": "(np.empty(0), np.empty(0))",
                    "doc": "Train and tests dataset splits.",
                    "typ": "Union[Tuple[tf.data.Dataset, "
                    "tf.data.Dataset], "
                    "Tuple[np.ndarray, "
                    "np.ndarray]]",
                },
            ),
        )
    ),
}

intermediate_repr_no_default_sql_doc = deepcopy(intermediate_repr_no_default_doc)
intermediate_repr_no_default_sql_doc["params"]["dataset_name"][
    "doc"
] = "[PK] {}".format(
    intermediate_repr_no_default_sql_doc["params"]["dataset_name"]["doc"]
)

intermediate_repr_only_return_type = {
    "name": None,
    "type": "static",
    "doc": "Some comment",
    "params": OrderedDict(
        (("dataset_name", {"doc": "Example: foo"}),),
    ),
    "returns": OrderedDict(
        (
            (
                "return_type",
                {
                    "doc": "Train and tests dataset splits.",
                    "typ": "Union[Tuple[tf.data.Dataset, tf.data.Dataset], "
                    "Tuple[np.ndarray, np.ndarray]]",
                },
            ),
        ),
    ),
}

intermediate_repr = {
    "name": None,
    "type": "static",
    "doc": docstring_header_str.replace("\n", ""),
    "params": OrderedDict(
        (
            (
                "dataset_name",
                {
                    "default": "mnist",
                    "doc": "name of dataset. Defaults to " '"mnist"',
                    "typ": "str",
                },
            ),
            (
                "tfds_dir",
                {
                    "default": "~/tensorflow_datasets",
                    "doc": "directory to look for models in. "
                    "Defaults to "
                    '"~/tensorflow_datasets"',
                    "typ": "str",
                },
            ),
            (
                "K",
                {
                    "default": "np",
                    "doc": "backend engine, e.g., `np` or " '`tf`. Defaults to "np"',
                    "typ": "Literal['np', 'tf']",
                },
            ),
            (
                "as_numpy",
                {
                    "default": NoneStr,
                    "doc": "Convert to numpy ndarrays.",
                    "typ": "Optional[bool]",
                },
            ),
            (
                "data_loader_kwargs",
                {
                    "default": NoneStr,
                    "doc": "pass this as arguments to " "data_loader function",
                    "typ": "Optional[dict]",
                },
            ),
        )
    ),
    "returns": OrderedDict(
        (
            (
                "return_type",
                {
                    "default": "(np.empty(0), np.empty(0))",
                    "doc": "Train and tests dataset splits. "
                    "Defaults to (np.empty(0), "
                    "np.empty(0))",
                    "typ": "Union[Tuple[tf.data.Dataset, "
                    "tf.data.Dataset], "
                    "Tuple[np.ndarray, "
                    "np.ndarray]]",
                },
            ),
        )
    ),
}

intermediate_repr_no_default_doc_or_prop = remove_defaults_from_intermediate_repr(
    deepcopy(intermediate_repr), emit_default_prop=False
)

__all__ = [
    "class_google_tf_tensorboard_ir",
    "class_torch_nn_l1loss_ir",
    "class_torch_nn_one_cycle_lr_ir",
    "docstring_google_pytorch_lbfgs_ir",
    "docstring_google_tf_adadelta_function_ir",
    "docstring_google_tf_adadelta_ir",
    "docstring_google_tf_adam_ir",
    "docstring_google_tf_lambda_callback_ir",
    "docstring_google_tf_squared_hinge_ir",
    "function_adder_ir",
    "intermediate_repr",
    "intermediate_repr_extra_colons",
    "intermediate_repr_no_default_doc",
    "intermediate_repr_no_default_doc_or_prop",
    "intermediate_repr_no_default_sql_doc",
    "intermediate_repr_only_return_type",
    "method_complex_args_variety_ir",
]



""" Tests for CLI gen subparser (__main__.py) """
import os
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, run_cli_test, unittest_main


class TestCliGen(TestCase):
    """Test class for __main__.py"""

    def test_gen_fails(self) -> None:
        """Tests CLI interface failure cases"""
        run_cli_test(
            self,
            ["gen", "--wrong"],
            exit_code=2,
            output="the following arguments are required: --name-tpl, --input-mapping, --emit, --output-filename/-o\n",
        )

    def test_existent_file_fails(self) -> None:
        """Tests nonexistent file throws the right error"""
        with TemporaryDirectory() as tempdir:
            filename = os.path.join(
                tempdir,
                "delete_this_1{}".format(os.path.basename(__file__)),
            )
            open(filename, "a").close()

            run_cli_test(
                self,
                [
                    "gen",
                    "--name-tpl",
                    "{name}Config",
                    "--input-mapping",
                    "cdd.pure_utils.simple_types",
                    "--emit",
                    "class",
                    "--output-filename",
                    filename,
                ],
                exception=OSError,
                exit_code=2,
                output="File exists and this is a destructive operation. Delete/move {!r} then rerun.".format(
                    filename
                ),
            )

    def test_gen(self) -> None:
        """Tests CLI interface gets all the way to the gen call without error"""
        with TemporaryDirectory() as tempdir:
            output_filename = os.path.join(tempdir, "classes.py")

            with patch("cdd.__main__.gen", mock_function):
                run_cli_test(
                    self,
                    [
                        "gen",
                        "--name-tpl",
                        "{name}Config",
                        "--input-mapping",
                        "cdd.pure_utils.simple_types",
                        "--emit",
                        "class",
                        "--output-filename",
                        output_filename,
                    ],
                    exit_code=None,
                    output=None,
                )


unittest_main()



"""
Docstring parsers.

Translates from the [ReST docstring format (Sphinx)](
  https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html#the-sphinx-docstring-format)

Translates from the [numpydoc docstring format](https://numpydoc.readthedocs.io/en/latest/format.html)

Translates from [Google's docstring format](https://google.github.io/styleguide/pyguide.html)
"""
import ast
from ast import AST
from collections import OrderedDict
from copy import deepcopy
from enum import Enum
from functools import partial
from itertools import chain, takewhile
from operator import attrgetter, contains, eq, le
from typing import Dict, List, Tuple

from cdd.ast_utils import NoneStr, get_value
from cdd.defaults_utils import _remove_default_from_param, needs_quoting
from cdd.docstring_utils import ARG_TOKENS, RETURN_TOKENS, TOKENS
from cdd.emit import to_code
from cdd.emitter_utils import interpolate_defaults
from cdd.pure_utils import (
    code_quoted,
    count_iter_items,
    identity,
    location_within,
    none_types,
    paren_wrap_code,
    rpartial,
    unquote,
    update_d,
)


class Style(Enum):
    """
    Simple enum taken from the docstring_parser codebase
    """

    rest = 1
    google = 2
    numpydoc = 3
    auto = 255


def parse_docstring(
    docstring,
    infer_type=False,
    default_search_announce=None,
    parse_original_whitespace=False,
    word_wrap=True,
    emit_default_prop=True,
    emit_default_doc=False,
):
    """Parse the docstring into its components.

    :param docstring: the docstring
    :type docstring: ```Optional[str]```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param emit_default_prop: Whether to include the default dictionary property.
    :type emit_default_prop: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :returns: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """

    assert isinstance(docstring, (type(None), str)), "{typ} != str".format(
        typ=type(docstring).__name__
    )
    style = derive_docstring_format(docstring)

    ir = {
        "name": None,
        "type": "static",
        # "_internal": {"_style": style.value},
        "doc": "",
        "params": OrderedDict(),
        "returns": None,
    }
    if not docstring:
        return ir

    scanned = _scan_phase(docstring, style=style)

    _parse_phase(
        ir,
        scanned,
        default_search_announce=default_search_announce,
        emit_default_doc=emit_default_doc,
        emit_default_prop=emit_default_prop,
        infer_type=infer_type,
        parse_original_whitespace=parse_original_whitespace,
        style=style,
        word_wrap=word_wrap,
    )

    # Apply certain functions regardless of style
    if style is Style.rest:
        ir.update(
            {
                k: OrderedDict(
                    map(
                        partial(
                            interpolate_defaults, emit_default_doc=emit_default_doc
                        ),
                        ir[k].items(),
                    )
                )
                if ir[k]
                else ir[k]
                for k in ("params", "returns")
            }
        )
    return ir


def derive_docstring_format(docstring):
    """
    Infer the docstring format of the provided docstring

    :param docstring: the docstring
    :type docstring: ```Optional[str]```

    :returns: the style of docstring
    :rtype: ```Literal['rest', 'numpydoc', 'google']```
    """
    if docstring is None or any(map(partial(contains, docstring), TOKENS.rest)):
        style = Style.rest
    elif any(map(partial(contains, docstring), TOKENS.google)):
        style = Style.google
    else:
        style = Style.numpydoc
    return style


def _scan_phase(docstring, parse_original_whitespace=False, style=Style.rest):
    """
    Scanner phase. Lexical analysis; to some degree…

    :param docstring: the docstring
    :type docstring: ```str```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param style: the style of docstring
    :type style: ```Style```

    :returns: List with each element a tuple of (whether value is a token, value)
    :rtype: ```Union[Dict[str, str], List[Tuple[bool, str]]]```
    """
    arg_tokens, return_tokens = map(attrgetter(style.name), (ARG_TOKENS, RETURN_TOKENS))
    return (
        _scan_phase_rest
        if style is Style.rest
        else partial(
            _scan_phase_numpydoc_and_google,
            parse_original_whitespace=parse_original_whitespace,
            style=style,
        )
    )(docstring, arg_tokens=arg_tokens, return_tokens=return_tokens)


def _scan_phase_numpydoc_and_google(
    docstring, parse_original_whitespace, arg_tokens, return_tokens, style
):
    """
    numpydoc and google scanner phase. Lexical analysis; to some degree…

    :param docstring: the docstring
    :type docstring: ```str```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param arg_tokens: Valid tokens like `"Parameters\n----------"`
    :type arg_tokens: ```Tuple[str]```

    :param return_tokens: Valid tokens like `"Returns\n-------"`
    :type return_tokens: ```Tuple[str]```

    :param style: the style of docstring
    :type style: ```Style```

    :returns: List with each element a tuple of (whether value is a token, value)
    :rtype: ```Dict[str, str]```
    """
    white_spacer = (
        identity
        if parse_original_whitespace
        else lambda s: s
        if s.isspace()
        else s.strip()
    )
    scanned: Dict[str, List[List[str]]] = {
        token: []
        for token in chain.from_iterable((("doc",), arg_tokens, return_tokens))
    }
    # ^ Dict[Union[Literal["doc"], arg_tokens, return_tokens], List[dict]]

    # First doc, if present
    _start_idx, _end_idx, _found = location_within(docstring, arg_tokens)
    # _leading_whitespace = "".join(takewhile(str.isspace, docstring[:_start_idx][::-1]))
    if _start_idx == -1:
        # Return type no args?
        _start_idx, _end_idx, _found = location_within(docstring, return_tokens)

    if _start_idx > -1:
        namespace = _found
        scanned["doc"] = white_spacer(docstring[:_start_idx])
        docstring = docstring[_end_idx + 1 :]
    else:
        scanned["doc"] = docstring
        return scanned

    # Scan all lines so that that each element in `stacker` refers to one 'unit'
    stacker, docstring_lines, line = [], docstring.splitlines(), None
    first_indent = count_iter_items(takewhile(str.isspace, docstring_lines[0]))
    for line_no, line in enumerate(docstring_lines):
        indent = count_iter_items(takewhile(str.isspace, line))

        if indent == first_indent:
            stacker.append([line])
        elif indent < first_indent:
            scanned[namespace] = scanned.get(namespace, []) + deepcopy(stacker)
            stacker.clear()
            if len(docstring_lines) > line_no + 3 and any(
                filter(rpartial(eq, docstring_lines[line_no + 1]), return_tokens)
                # return_token = return_tokens[0].splitlines()
                # filter(rpartial(eq, docstring_lines[line_no + 1]), return_tokens)
            ):
                return_indent = count_iter_items(
                    takewhile(str.isspace, docstring_lines[line_no + 3])
                )
                next_smallest_indent = count_iter_items(
                    takewhile(
                        partial(le, return_indent),
                        map(
                            lambda l: count_iter_items(takewhile(str.isspace, l)),
                            docstring_lines[line_no + 3 :],
                        ),
                    )
                )
                scanned[return_tokens[0]] = docstring_lines[
                    line_no + 2 : line_no + 3 + next_smallest_indent
                ]
                scanned_afterward = docstring_lines[
                    line_no + 3 + next_smallest_indent :
                ]
            else:
                scanned_afterward = docstring_lines[line_no + 1 :]
                if (
                    len(scanned_afterward) > 1
                    and scanned_afterward[0] == return_tokens[0]
                ):
                    return_indent = count_iter_items(
                        takewhile(str.isspace, scanned_afterward[1])
                    )
                    next_smallest_indent = count_iter_items(
                        takewhile(
                            partial(le, return_indent),
                            map(
                                lambda l: count_iter_items(takewhile(str.isspace, l)),
                                scanned_afterward[2:],
                            ),
                        )
                    )
                    scanned[return_tokens[0]] = scanned_afterward[
                        1 : next_smallest_indent + 2
                    ]
                    scanned_afterward = (
                        None
                        if next_smallest_indent == 0
                        else scanned_afterward[next_smallest_indent + 2 :]
                    )

            if scanned_afterward:
                scanned["scanned_afterward"] = scanned_afterward
            break
        else:
            stacker[-1].append(line)

    if line is not None and (not stacker or not stacker[-1] or stacker[-1][0] != line):
        if "scanned_afterward" in scanned:
            scanned["scanned_afterward"].insert(0, line)
        else:
            scanned["scanned_afterward"] = [line]

    # Split out return, if present and not already set
    if not scanned.get(return_tokens[0], False):
        stacker = _return_parse_phase_numpydoc_and_google(
            return_tokens, scanned, stacker, style
        )

    if stacker:
        scanned[namespace] = stacker

    return scanned


def _return_parse_phase_numpydoc_and_google(return_tokens, scanned, stacker, style):
    """
    numpydoc and google scanner phase for return. Lexical analysis; to some degree…

    :param return_tokens: Valid tokens like `"Returns\n-------"`
    :type return_tokens: ```Tuple[str]```

    :param scanned: List with each element a tuple of (whether value is a token, value)
    :type scanned: ```Union[Dict[str, str], List[Tuple[bool, str]]]```

    :param stacker: Stack of strings part that forms int scanned
    :type stacker: ```List[List[str]]```

    :param style: the style of docstring
    :type style: ```Style```

    :returns: Whatever is left of `stacker`. This function may also set the return key of the `scanned`
    :rtype: ```List[List[str]]```
    """
    rev_return_token = return_tokens[0].splitlines()[::-1]
    rng = range(len(stacker) - 1, -1, -1)
    if style is Style.numpydoc:
        for i in rng:
            if i - 1 > 0 and stacker[i] + stacker[i - 1] == rev_return_token:
                scanned[return_tokens[0]] = stacker[i + 1 :]
                stacker = stacker[: i - 1]
                break

    return stacker


def _scan_phase_rest(docstring, arg_tokens, return_tokens):
    """
    Scanner phase. Lexical analysis; to some degree…

    :param docstring: the docstring
    :type docstring: ```str```

    :param arg_tokens: Valid tokens like `":param"`
    :type arg_tokens: ```Tuple[str]```

    :param return_tokens: Valid tokens like `":rtype:"`
    :type return_tokens: ```Tuple[str]```

    :returns: List with each element a tuple of (whether value is a token, value)
    :rtype: ```List[Tuple[bool, str]]```
    """

    all_tokens = arg_tokens + return_tokens
    rev_known_tokens_t = tuple(map(tuple, map(reversed, all_tokens)))
    scanned: List[Tuple[bool, str]] = []
    stack: List[str] = []

    for ch in docstring:
        stack.append(ch)

        stack_rev = stack[::-1]

        for token in rev_known_tokens_t:
            token_len = len(token)
            if tuple(stack_rev[:token_len]) == token:
                scanned.append((bool(len(scanned)), "".join(stack[:-token_len])))
                stack = stack[len(scanned[-1][1]) :][:token_len]
                continue

    if stack:
        final = "".join(stack)
        scanned.append(
            (
                bool(scanned and scanned[-1][0])
                or any(map(final.startswith, all_tokens)),
                final,
            )
        )

    return scanned


def _parse_phase(
    intermediate_repr,
    scanned,
    default_search_announce,
    infer_type,
    parse_original_whitespace,
    word_wrap,
    emit_default_prop,
    emit_default_doc,
    style=Style.rest,
):
    """
    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param scanned: List with each element a tuple of (whether value is a token, value)
    :type scanned: ```Union[Dict[str, str], List[Tuple[bool, str]]]```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param emit_default_prop: Whether to include the default dictionary property.
    :type emit_default_prop: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param style: the style of docstring
    :type style: ```Style```
    """
    arg_tokens, return_tokens = map(attrgetter(style.name), (ARG_TOKENS, RETURN_TOKENS))
    (
        _parse_phase_rest
        if style is Style.rest
        else partial(
            _parse_phase_numpydoc_and_google, style=style, arg_tokens=arg_tokens
        )
    )(
        intermediate_repr,
        scanned,
        default_search_announce=default_search_announce,
        emit_default_doc=emit_default_doc,
        emit_default_prop=emit_default_prop,
        infer_type=infer_type,
        parse_original_whitespace=parse_original_whitespace,
        return_tokens=return_tokens,
        word_wrap=word_wrap,
    )


def _set_name_and_type(param, infer_type, word_wrap):
    """
    Sanitise the name and set the type (iff default and no existing type) for the param

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :returns: Name, dict with keys: 'typ', 'doc', 'default'
    :rtype: ```Tuple[str, dict]```
    """
    name, _param = param
    del param
    if name is not None and (name.endswith("kwargs") or name.startswith("**")):
        name = name.lstrip("*")
        if _param.get("typ", "dict") == "dict":
            _param["typ"] = "Optional[dict]"
        if "default" not in _param:
            _param["default"] = NoneStr
    elif "default" in _param:
        _infer_default(_param, infer_type)
    google_opt = ", optional"
    if _param.get("typ"):
        _param["typ"] = (
            "Optional[{}]".format(_param["typ"][: -len(google_opt)])
            if _param["typ"].endswith(google_opt)
            else _param["typ"]
        )
    if "doc" in _param and not _param["doc"]:
        del _param["doc"]

    # if "doc" in _param and isinstance(_param["doc"], list):
    #     _param["doc"] = "".join(_param["doc"])
    if "doc" in _param:
        if not isinstance(_param["doc"], str):
            _param["doc"] = "".join(_param["doc"]).rstrip()
        else:
            _param["doc"] = (
                " ".join(map(str.strip, _param["doc"].split("\n")))
                if word_wrap
                else _param["doc"]
            ).rstrip()
        if (
            (
                _param["doc"].startswith("(Optional)")
                or _param["doc"].startswith("Optional")
            )
            and "typ" in _param
            and not _param["typ"].startswith("Optional[")
        ):
            _param["typ"] = "Optional[{}]".format(_param["typ"])

    return name, _param


def _infer_default(_param, infer_type):
    """
    Internal function to infer the default. Not intended for use by more than [the current] one function.

    :param _param: dict with keys: 'typ', 'doc', 'default'
    :type _param: ```dict```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```
    """
    if isinstance(
        _param["default"], (ast.Str, ast.Num, ast.Constant, ast.NameConstant)
    ):
        _param["default"] = get_value(_param["default"])
    if _param.get("default", False) in none_types:
        _param["default"] = NoneStr
    if infer_type and _param.get("typ") is None and _param["default"] not in none_types:
        _param["typ"] = type(_param["default"]).__name__
    if needs_quoting(_param.get("typ")) or isinstance(_param["default"], str):
        _param["default"] = unquote(_param["default"])
    elif isinstance(_param["default"], AST):
        try:
            _param["default"] = ast.literal_eval(_param["default"])
            # if _param.get("typ") is None or _param["typ"] == "UnaryOp":
            #    _param["typ"] = type(_param["default"]).__name__
        except ValueError:
            _param["default"] = "```{default}```".format(
                default=paren_wrap_code(to_code(_param["default"]).rstrip("\n"))
            )
    if _param.get("typ") is None and _param["default"] != NoneStr:
        _param["typ"] = type(_param["default"]).__name__
    if (
        _param["default"] != NoneStr
        and code_quoted(_param["default"])
        and "["
        not in _param.get(
            "typ", iter(())
        )  # Skip if you've actually formed a proper type
    ):
        del _param["typ"]  # Could make it `object` I suppose…


def _parse_phase_numpydoc_and_google(
    intermediate_repr,
    scanned,
    default_search_announce,
    infer_type,
    word_wrap,
    style,
    arg_tokens,
    return_tokens,
    parse_original_whitespace,
    emit_default_prop,
    emit_default_doc,
):
    """
    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param scanned: List with each element a tuple of (whether value is a token, value)
    :type scanned: ```Dict[str, str]```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param style: the style of docstring
    :type style: ```Style```

    :param arg_tokens: Valid tokens like `"Parameters\n----------"`
    :type arg_tokens: ```Tuple[str]```

    :param return_tokens: Valid tokens like `"Returns\n-------"`
    :type return_tokens: ```Tuple[str]```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param emit_default_prop: Whether to include the default dictionary property.
    :type emit_default_prop: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```
    """
    white_spacer = identity if parse_original_whitespace else str.lstrip
    if style is Style.numpydoc:

        def _parse(scan):
            """
            Parse the scanned input (numpydoc)

            :param scan: Scanned input
            :type scan: ```List[str]```

            :returns: dict with keys: 'name', 'typ', 'doc'
            :rtype: ```dict```
            """
            name, _, typ = scan[0].partition(":")
            if not name:
                return None
            cur = {"name": name.strip()}
            if typ:
                cur.update(
                    {"typ": typ.lstrip(), "doc": "\n".join(map(white_spacer, scan[1:]))}
                )
            return cur

    else:

        def _parse(scan, partitioned=None):
            """
            Parse the scanned input (Google)

            :param scan: Scanned input
            :type scan: ```List[str]```

            :param partitioned: Prep-partitioned `scan`, if given doesn't partition on `scan`, just uses this
            :type partitioned: ```Optional[Tuple[str, str, str]]```

            :returns: dict of shape {'name': ..., 'typ': ..., 'doc': ..., 'default': ..., 'required': ... }
            :rtype: ```dict```
            """
            offset = next(idx for idx, ch in enumerate(scan[0]) if ch == ":")
            s = white_spacer(scan[0][:offset])
            name, delim, typ = partitioned or s.partition("(")
            name, typ = name.strip(), (delim + typ).rstrip()
            cur = {"name": name}
            if typ:
                assert typ.startswith("(") and typ.endswith(
                    ")"
                ), "Expected third partition to be paren wrapped {!r}".format(s)
                cur["typ"] = typ[1:-1]
                if " or " in cur["typ"]:
                    cur["typ"] = "Union[{}]".format(", ".join(cur["typ"].split(" or ")))
                end = white_spacer(scan[0][offset + 1 :])
                if len(end) > 3 and end.startswith("{") and end.endswith("}"):
                    # PyTorch invented their own syntax for this I guess?
                    cur["typ"], scan[0] = (
                        "Literal{}".format(
                            list(map(rpartial(str.strip, "'"), end[1:-1].split(", ")))
                        ),
                        "",
                    )
            cur["doc"] = (lambda s_: s_ if parse_original_whitespace else s_.strip())(
                "\n".join(
                    chain.from_iterable(
                        ((white_spacer(scan[0][offset + 1 :]),), scan[1:])
                    )
                )
            )
            return cur

    scanned_params = scanned[arg_tokens[0]]

    # Handle stuff after the Args, e.g., usage notes; doctests; references.
    afterward_idx = next(
        (
            idx
            for idx, elem in enumerate(scanned_params)
            if elem[0].endswith(":") and elem[0].count(":") == 1
        ),
        -1,
    )
    if afterward_idx > -1:
        scanned_params, scanned_afterward = (
            scanned_params[:afterward_idx],
            scanned_params[afterward_idx:],
        )
        scanned["doc"] += "\n\n{}".format("\n".join(map("\n".join, scanned_afterward)))

    if "scanned_afterward" in scanned:
        _fill_doc_with_afterward(scanned)

    def _interpolate_defaults_and_force_future_default(name_param):
        """
        interpolate the defaults and force future default if current default is set

        :param name_param: Name, dict with keys: 'typ', 'doc', 'default'
        :type name_param: ```Tuple[str, dict]```

        :returns: Name, dict with keys: 'typ', 'doc', 'default'
        :rtype: ```Tuple[str, dict]```
        """
        name, param = interpolate_defaults(
            name_param,
            emit_default_doc=emit_default_doc,
            require_default=_interpolate_defaults_and_force_future_default.require_default,
            default_search_announce=default_search_announce,
        )
        if (
            not _interpolate_defaults_and_force_future_default.require_default
            and param.get("default") is not None
        ):
            _interpolate_defaults_and_force_future_default.require_default = True
        return name, param

    _interpolate_defaults_and_force_future_default.require_default = False

    intermediate_repr.update(
        {
            "doc": "" if scanned["doc"].isspace() else white_spacer(scanned["doc"]),
            "params": OrderedDict(
                map(
                    partial(
                        _set_name_and_type, infer_type=infer_type, word_wrap=word_wrap
                    ),
                    map(
                        _interpolate_defaults_and_force_future_default,
                        map(
                            lambda d: (d.pop("name"), d),
                            filter(None, map(_parse, scanned_params)),
                        ),
                    ),
                ),
            ),
            "returns": OrderedDict(
                (
                    _interpolate_defaults_and_force_future_default(
                        _set_name_and_type(
                            (
                                "return_type",
                                (
                                    {
                                        "typ": scanned[return_tokens[0]][0][
                                            :-1
                                        ].lstrip(),
                                        "doc": white_spacer(
                                            scanned[return_tokens[0]][1]
                                        ),
                                    }
                                    if len(scanned[return_tokens[0]]) == 2
                                    and isinstance(scanned[return_tokens[0]][1], str)
                                    else {}
                                    if isinstance(scanned[return_tokens[0]][0], str)
                                    and scanned[return_tokens[0]][0].isspace()
                                    else {
                                        "doc": white_spacer(
                                            scanned[return_tokens[0]][0]
                                        )
                                        if isinstance(scanned[return_tokens[0]][0], str)
                                        else scanned[return_tokens[0]][0]
                                    }
                                )
                                if style is Style.google
                                else {
                                    "typ": scanned[return_tokens[0]][0][0],
                                    "doc": white_spacer(
                                        scanned[return_tokens[0]][0][1]
                                    ),
                                },
                            ),
                            infer_type=infer_type,
                            word_wrap=word_wrap,
                        ),
                    ),
                ),
            )
            if scanned[return_tokens[0]]
            else None,
        }
    )


def _fill_doc_with_afterward(scanned):
    """
    Internal func for `_parse_phase_numpydoc_and_google` to parse then add `scanned_afterward` to `doc` of `scanned`

    :param scanned: Dict with `scanned_afterward` and `doc` properties
    :type scanned: ```Dict[str, str]```
    """

    if next(
        filter(
            partial(eq, scanned["scanned_afterward"][0]),
            (
                e
                for k, v in scanned.items()
                for elem in v
                for e in elem
                if k != "scanned_afterward"
            ),
        ),
        False,
    ):
        del scanned["scanned_afterward"]
    else:
        scanned["doc"] += (
            "\n"
            if scanned["scanned_afterward"] and scanned["scanned_afterward"][0] == ""
            else ""
        ) + "\n".join(scanned["scanned_afterward"])


def _parse_phase_rest(
    intermediate_repr,
    scanned,
    default_search_announce,
    infer_type,
    word_wrap,
    parse_original_whitespace,
    emit_default_prop,
    emit_default_doc,
    return_tokens,
):
    """
    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param scanned: List with each element a tuple of (whether value is a token, value)
    :type scanned: ```List[Tuple[bool, str]]```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param emit_default_prop: Whether to include the default dictionary property.
    :type emit_default_prop: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param return_tokens: Valid tokens like `":rtype:"`
    :type return_tokens: ```Tuple[str]```
    """
    param = [
        None,
        {},
    ]  # First elem is name and second elem is dict with keys: 'typ', 'doc', 'default'
    for is_token, line in scanned:
        if is_token is True:
            if any(map(line.startswith, return_tokens)):
                nxt_colon = line.find(":", 1)
                val = line[nxt_colon + 1 :].strip()
                if intermediate_repr["returns"] is None:
                    intermediate_repr["returns"] = OrderedDict((("return_type", {}),))
                intermediate_repr["returns"]["return_type"].update(
                    interpolate_defaults(
                        (
                            "return_type",
                            dict((_set_param_values(line, val, return_tokens[-1]),)),
                        ),
                        emit_default_doc=emit_default_doc,
                        default_search_announce=default_search_announce,
                    )[1]
                )
            else:
                fst_space = line.find(" ")
                nxt_colon = line.find(":", fst_space)
                name = line[fst_space + 1 : nxt_colon]

                if param[0] is not None and not param[0] == name:
                    if not param[0][0] == "*":
                        intermediate_repr["params"][param[0]] = param[1]
                    param = [None, {}]

                val = (lambda s_: s_ if parse_original_whitespace else s_.strip())(
                    line[nxt_colon + 1 :]
                )

                param = _set_name_and_type(
                    interpolate_defaults(
                        (
                            name,
                            update_d(param[1], dict((_set_param_values(line, val),))),
                        ),
                        emit_default_doc=emit_default_doc,
                    ),
                    infer_type=infer_type,
                    word_wrap=word_wrap,
                )
                if not emit_default_doc and not emit_default_prop:
                    param = _remove_default_from_param(
                        param, emit_default_prop=emit_default_doc
                    )
        elif not intermediate_repr["doc"]:
            intermediate_repr["doc"] = (
                line if parse_original_whitespace else line.strip()
            )
    if param != [None, {}]:
        # if param['name'] == 'return_type': intermediate_repr['returns'] = param
        name, param = _set_name_and_type(
            interpolate_defaults(param, emit_default_doc=emit_default_doc),
            infer_type=infer_type,
            word_wrap=word_wrap,
        )
        if not emit_default_doc and not emit_default_prop:
            name, param = _remove_default_from_param(
                (name, param), emit_default_prop=emit_default_doc
            )
        intermediate_repr["params"][name] = param


def _set_param_values(input_str, val, sw=":type"):
    """
    Sets the typ or doc values properly.

    :param val: The value (`sw` figures out what it means semantically)
    :type val: ```str```

    :param sw: Startswith condition
    :type sw: ```str```

    :returns: Properly derived key and [potentially modified] value
    :rtype: Tuple[Literal['doc', 'typ'], str]
    """
    return (
        ("typ", (lambda v: "dict" if v.startswith("**") else v)(val.replace("```", "")))
        if input_str.startswith(sw)
        else ("doc", val)
    )


__all__ = ["derive_docstring_format", "parse_docstring"]



""" Tests for emitter_utils """
from ast import Attribute, Call, Expr, Load, Name, Subscript, keyword
from copy import deepcopy
from unittest import TestCase

from cdd.ast_utils import set_value
from cdd.emitter_utils import interpolate_defaults, parse_out_param
from cdd.pure_utils import rpartial
from cdd.tests.mocks.argparse import argparse_add_argument_ast, argparse_func_ast
from cdd.tests.mocks.ir import intermediate_repr
from cdd.tests.utils_for_tests import unittest_main


class TestEmitterUtils(TestCase):
    """Test class for emitter_utils"""

    def test_parse_out_param(self) -> None:
        """Test that parse_out_param parses out the right dict"""
        self.assertDictEqual(
            parse_out_param(
                next(filter(rpartial(isinstance, Expr), argparse_func_ast.body[::-1]))
            )[1],
            # Last element:
            intermediate_repr["params"]["data_loader_kwargs"],
        )

    def test_parse_out_param_default(self) -> None:
        """Test that parse_out_param sets default when required and unset"""

        self.assertDictEqual(
            parse_out_param(argparse_add_argument_ast)[1],
            {"default": 0, "doc": None, "typ": "int"},
        )

    def test_parse_out_param_fails(self) -> None:
        """Test that parse_out_param throws NotImplementedError when unsupported type given"""
        self.assertRaises(
            NotImplementedError,
            lambda: parse_out_param(
                Expr(
                    Call(
                        args=[set_value("--num")],
                        func=Attribute(
                            Name("argument_parser", Load()),
                            "add_argument",
                            Load(),
                        ),
                        keywords=[
                            keyword(
                                arg="type",
                                value=Subscript(
                                    expr_context_ctx=None,
                                    expr_slice=None,
                                    expr_value=None,
                                ),
                                identifier=None,
                            ),
                            keyword(
                                arg="required",
                                value=set_value(True),
                                identifier=None,
                            ),
                        ],
                        expr=None,
                        expr_func=None,
                    )
                )
            ),
        )

    def test_interpolate_defaults(self) -> None:
        """Test that interpolate_defaults corrects sets the default property"""
        param = "K", deepcopy(intermediate_repr["params"]["K"])
        param_with_correct_default = deepcopy(param[1])
        del param[1]["default"]
        self.assertDictEqual(interpolate_defaults(param)[1], param_with_correct_default)


unittest_main()



"""
Shared utility functions used by many tests
"""
import ast
from copy import deepcopy
from functools import partial
from importlib import import_module
from importlib.abc import Loader
from importlib.util import module_from_spec, spec_from_loader
from itertools import takewhile
from os import path
from sys import modules
from tempfile import NamedTemporaryFile
from unittest import main
from unittest.mock import MagicMock, patch

from cdd import source_transformer
from cdd.ast_utils import cmp_ast, set_value
from cdd.docstring_utils import TOKENS
from cdd.pure_utils import PY3_8, count_iter_items, identity, reindent, tab

black = (
    import_module("black")
    if "black" in modules
    else type(
        "black",
        tuple(),
        {
            "format_str": lambda src_contents, mode: None,
            "Mode": lambda target_versions, line_length, is_pyi, string_normalization: None,
        },
    )
)


def run_ast_test(test_case_instance, gen_ast, gold, skip_black=False):
    """
    Compares `gen_ast` with `gold` standard

    :param test_case_instance: instance of `TestCase`
    :type test_case_instance: ```unittest.TestCase```

    :param gen_ast: generated AST
    :type gen_ast: ```Union[ast.Module, ast.ClassDef, ast.FunctionDef]```

    :param skip_black: Whether to skip formatting with black. Turned off for performance, turn on for pretty debug.
    :type skip_black: ```bool```

    :param gold: mocked AST
    :type gold: ```Union[ast.Module, ast.ClassDef, ast.FunctionDef]```
    """
    if isinstance(gen_ast, str):
        gen_ast = ast.parse(gen_ast).body[0]

    assert gen_ast is not None, "gen_ast is None"
    assert gold is not None, "gold is None"

    gen_ast = deepcopy(gen_ast)
    gold = deepcopy(gold)

    # if reindent_docstring:
    #     gen_docstring = ast.get_docstring(gen_ast)
    #     if gen_docstring is not None:
    #         gen_ast.body[0] = set_value(
    #             "\n{}".format(indent(cleandoc(gen_docstring), tab))
    #         )
    #     gold.body[0] = set_value(
    #       "\n{}".format(indent(ast.get_docstring(gold, clean=True), tab))
    #     )

    # from meta.asttools import print_ast
    #
    # print("#gen")
    # print_ast(gen_ast)
    # print("#gold")
    # print_ast(gold)

    _gen_ast, _gold_ast = (
        (gen_ast.body[0], gold.body[0])
        if isinstance(gen_ast, ast.Module) and gen_ast.body
        else (gen_ast, gold)
    )
    if isinstance(_gen_ast, (ast.ClassDef, ast.AsyncFunctionDef, ast.FunctionDef)):
        test_case_instance.assertEqual(
            *map(partial(ast.get_docstring, clean=False), (_gen_ast, _gold_ast))
        )

    test_case_instance.assertEqual(
        *map(
            identity
            if skip_black
            else partial(
                black.format_str,
                mode=black.Mode(
                    target_versions=set(),
                    line_length=60,
                    is_pyi=False,
                    string_normalization=False,
                ),
            ),
            map(source_transformer.to_code, (gen_ast, gold)),
        )
    )

    test_case_instance.assertTrue(
        cmp_ast(gen_ast, gold), "Generated AST doesn't match reference AST"
    )


def run_cli_test(
    test_case_instance,
    cli_argv,
    exit_code,
    output,
    output_checker=lambda output: (lambda q: output[output.find(q) + len(q) :])(
        "error: "
    ),
    exception=SystemExit,
    return_args=False,
):
    """
    CLI test helper, wraps exit code and stdout/stderr input_str

    :param test_case_instance: instance of `TestCase`
    :type test_case_instance: ```unittest.TestCase```

    :param cli_argv: cli_argv, can be sys.argv or proxy
    :type cli_argv: ```List[str]```

    :param exit_code: exit code
    :type exit_code: ```Optional[int]```

    :param output: string representation (from stdout/stderr)
    :type output: ```Optional[str]```

    :param output_checker: Function to check the input_str with
    :type output_checker: ```Callable[[str], bool]```

    :param exception: The exception that is expected to be raised
    :type exception: ```Union[BaseException, Exception]```

    :param return_args: Primarily use is for tests. Returns the args rather than executing anything.
    :type return_args: ```bool```

    :returns: input_str
    :rtype: ```Tuple[str, Optional[Namespace]]```
    """
    argparse_mock, args = MagicMock(), None
    with patch("argparse.ArgumentParser._print_message", argparse_mock), patch(
        "sys.argv", cli_argv
    ):
        from cdd.__main__ import main

        main_f = partial(main, cli_argv=cli_argv, return_args=return_args)
        if exit_code is None:
            args = main_f()
        else:
            with test_case_instance.assertRaises(exception) as e:
                args = main_f()
    if exit_code is not None:
        test_case_instance.assertEqual(
            *(e.exception.code, exception(exit_code).code)
            if exception is SystemExit
            else (str(e.exception), output)
        )
    if exception is not SystemExit:
        pass
    elif argparse_mock.call_args is None:
        test_case_instance.assertIsNone(output)
    else:
        test_case_instance.assertEqual(
            output_checker(
                (argparse_mock.call_args.args if PY3_8 else argparse_mock.call_args[0])[
                    0
                ]
            ),
            output,
        )
    return output, args


def unittest_main():
    """Runs unittest.main if __main__"""
    if __name__ == "__main__":
        main()


# These next two are mostly https://stackoverflow.com/a/64960448


class ShowSourceLoader(Loader):
    """
    Loader that will enable `inspect.getsource` and friends to work from an in-memory construct
    """

    def __init__(self, modname, source):
        """
        :param modname: Name of module
        :type modname: ```str```

        :param source: Source string
        :type source: ```str```
        """
        self.modname = modname
        self.source = source

    def get_source(self, modname):
        """
        Return the source for the module

        :param modname: Name of module
        :type modname: ```str```

        :returns: Source string
        :rtype: ```str```
        """
        assert modname == self.modname, ImportError(modname)
        return self.source


def inspectable_compile(s, modname=None):
    """
    Compile and executable the input string, returning what was constructed

    :param s: Input source
    :type s: ```str```

    :param modname: Module name, generates a random one if None
    :type modname: ```Optional[str]```

    :returns: The compiled and executed input source module, such that `inspect.getsource` works
    :rtype: ```Any```
    """
    fh = NamedTemporaryFile(suffix=".py")
    filename = fh.name
    try:
        modname = modname or path.splitext(path.basename(filename))[0]
        assert modname not in modules
        # our loader is a dummy one which just spits out our source
        loader = ShowSourceLoader(modname, s)
        spec = spec_from_loader(modname, loader, origin=filename)
        module = module_from_spec(spec)
        # the code must be compiled so the function's code object has a filename
        code = compile(s, mode="exec", filename=filename)
        exec(code, module.__dict__)
        # inspect.getmodule(...) requires it to be in sys.modules
        setattr(module, "__file__", s)
        modules[modname] = module
        return module
    finally:
        fh.close()  # Is auto-deleted on close


def mock_function(*args, **kwargs):
    """
    Mock function to check if it is called

    :returns: True
    :rtype: ```Literal[True]```
    """
    return True


def reindent_docstring(node, indent_level=1):
    """
    Reindent the docstring

    :param node: AST node
    :type node: ```ast.AST```

    :param indent_level: docstring indentation level whence: 0=no_tabs, 1=one tab; 2=two tabs
    :type indent_level: ```int```

    :returns: Node with reindent docstring
    :rtype: ```ast.AST```
    """
    doc_str = ast.get_docstring(node)
    if doc_str is not None:
        node.body[0] = ast.Expr(
            set_value(
                "\n{tab}{s}\n{tab}".format(
                    tab=tab * abs(indent_level),
                    s="\n".join(
                        map(
                            lambda line: "{sep}{line}".format(
                                sep=tab * 2, line=line.lstrip()
                            )
                            if line.startswith(tab)
                            and len(line) > len(tab)
                            and line[
                                len(tab) : line.lstrip().find(" ") + len(tab)
                            ].rstrip(":s")
                            not in frozenset((False,) + TOKENS.rest)
                            else line,
                            reindent(doc_str).splitlines(),
                        )
                    ),
                )
            )
        )
    return node


def remove_args_from_docstring(doc_str):
    """
    Remove args, kwargs, raises, and any other "args" from the docstring

    :param doc_str: The doc str (any style)
    :type doc_str: ```str```

    :returns: Docstring excluding args
    :rtype: ```str```
    """
    stack, in_args = [], False
    assert isinstance(doc_str, str)
    for line in doc_str.splitlines():
        stripped_line = line.lstrip()
        if any(filter(stripped_line.startswith, TOKENS)):
            in_args = True
        elif line.isspace():
            stack.append("")
        elif (
            not in_args
            or stripped_line.endswith(":")
            and count_iter_items(takewhile(str.isalpha, stripped_line[:-1]))
            == len(stripped_line) - 1
        ):
            stack.append(line)
            in_args = False
    return "{}{}".format("\n".join(stack), "")  # ("\n" if stack[-1] == "" else "")


__all__ = [
    "inspectable_compile",
    "mock_function",
    "remove_args_from_docstring",
    "run_ast_test",
    "run_cli_test",
    "unittest_main",
]



"""
Tests OpenAPI
"""
from unittest import TestCase

from cdd.openapi.emit import openapi
from cdd.openapi.emitter_utils import NameModelRouteIdCrud
from cdd.tests.mocks.json_schema import config_schema
from cdd.tests.mocks.openapi import openapi_dict
from cdd.tests.mocks.routes import route_config
from cdd.tests.utils_for_tests import unittest_main


class TestOpenApi(TestCase):
    """Tests whether `NameModelRouteIdCrud` can construct a `dict`"""

    def test_openapi_emitter(self) -> None:
        """
        Tests whether `openapi.emit` produces `openapi_dict` given `NameModelRouteIdCrud`
        """
        self.assertDictEqual(
            openapi(
                (
                    NameModelRouteIdCrud(
                        name=route_config["name"],
                        model=config_schema,
                        route=route_config["route"],
                        id=route_config["primary_key"],
                        crud="CRD",
                    ),
                )
            ),
            openapi_dict,
        )


unittest_main()



"""
Tests for docstring parsing
"""

from ast import BinOp, Mult
from collections import OrderedDict
from copy import deepcopy
from unittest import TestCase

import cdd.emit
import cdd.emitter_utils
from cdd import parse
from cdd.ast_utils import set_value
from cdd.docstring_parsers import _set_name_and_type, parse_docstring
from cdd.tests.mocks.docstrings import (
    docstring_extra_colons_str,
    docstring_google_pytorch_lbfgs_str,
    docstring_google_str,
    docstring_google_tf_adadelta_str,
    docstring_google_tf_adam_str,
    docstring_google_tf_lambda_callback_str,
    docstring_google_tf_squared_hinge_str,
    docstring_header_and_return_str,
    docstring_header_str,
    docstring_no_default_doc_str,
    docstring_numpydoc_only_doc_str,
    docstring_numpydoc_only_params_str,
    docstring_numpydoc_only_returns_str,
    docstring_numpydoc_str,
    docstring_only_return_type_str,
    docstring_str,
)
from cdd.tests.mocks.ir import (
    docstring_google_pytorch_lbfgs_ir,
    docstring_google_tf_adadelta_ir,
    docstring_google_tf_adam_ir,
    docstring_google_tf_lambda_callback_ir,
    docstring_google_tf_squared_hinge_ir,
    intermediate_repr,
    intermediate_repr_extra_colons,
    intermediate_repr_no_default_doc,
    intermediate_repr_no_default_doc_or_prop,
    intermediate_repr_only_return_type,
)
from cdd.tests.utils_for_tests import unittest_main


class TestMarshallDocstring(TestCase):
    """
    Tests whether docstrings are parsed out—and emitted—correctly
    """

    def test_ir_equality(self) -> None:
        """
        Tests whether `parse_docstring` produces `intermediate_repr`
              from `docstring_str`"""
        self.assertDictEqual(
            parse_docstring(docstring_str), intermediate_repr_no_default_doc
        )

    def test_intermediate_repr_no_default_doc_equality(self) -> None:
        """
        Tests whether `parse_docstring` produces `docstring_str_no_default_doc`
              from `docstring_str_no_default_doc`"""
        ir = parse_docstring(
            docstring_no_default_doc_str,
            emit_default_doc=False,
            emit_default_prop=False,
        )
        self.assertDictEqual(
            ir,
            intermediate_repr_no_default_doc_or_prop,
        )

    def test_intermediate_repr_extra_colons_equality(self) -> None:
        """
        Tests whether `parse_docstring` produces `docstring_str_no_default_doc`
              from `docstring_str_no_default_doc`"""

        self.assertDictEqual(
            parse_docstring(docstring_extra_colons_str, emit_default_doc=False),
            intermediate_repr_extra_colons,
        )

    def test_intermediate_repr_only_return_type_equality(self) -> None:
        """
        Tests whether `parse_docstring` produces `docstring_str_no_default_doc`
              from `docstring_str_no_default_doc`"""

        self.assertDictEqual(
            parse_docstring(docstring_only_return_type_str, emit_default_doc=False),
            intermediate_repr_only_return_type,
        )

    def test_ir2docstring(self) -> None:
        """Tests whether `emit.docstring` produces `docstring_str` from `intermediate_repr`"""
        self.assertEqual(
            docstring_str.strip(),
            cdd.emit.docstring(
                deepcopy(intermediate_repr),
                indent_level=0,
                emit_types=True,
                emit_default_doc=True,
                emit_separating_tab=False,
                word_wrap=False,
            ).strip(),
        )

    def test_from_docstring(self) -> None:
        """
        Tests whether `docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_str`"""
        ir, returns = parse.docstring(
            docstring_str, return_tuple=True, emit_default_doc=False
        )
        self.assertTrue(returns)
        self.assertDictEqual(ir, intermediate_repr_no_default_doc)

    def test_from_docstring_numpydoc(self) -> None:
        """
        Tests whether `docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_numpydoc_str`"""
        ir, returns = parse.docstring(
            docstring_numpydoc_str, return_tuple=True, emit_default_doc=False
        )
        self.assertTrue(returns)
        self.assertDictEqual(ir, intermediate_repr_no_default_doc)

    def test_from_docstring_numpydoc_only_params(self) -> None:
        """
        Tests whether `docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_numpydoc_only_params_str`"""
        ir, returns = parse.docstring(
            docstring_numpydoc_only_params_str,
            return_tuple=True,
            emit_default_doc=False,
        )
        self.assertFalse(returns)
        gold = deepcopy(intermediate_repr_no_default_doc)
        del gold["returns"]
        gold.update({"doc": "", "returns": None})
        self.assertDictEqual(ir, gold)

    def test__set_name_and_type(self) -> None:
        """
        Tests that `_set_name_and_type` parsed AST code into a code str.
        Not working since I explicitly deleted the typ from ``` quoted defaults. Changed mock to match.
        """
        self.assertTupleEqual(
            _set_name_and_type(
                (
                    "adder",
                    {
                        "default": BinOp(
                            set_value(5),
                            Mult(),
                            set_value(5),
                        ),
                    },
                ),
                infer_type=True,
                word_wrap=True,
            ),
            ("adder", {"default": "```(5 * 5)```"}),
        )

        self.assertTupleEqual(
            _set_name_and_type(
                (
                    "adder",
                    {
                        "default": BinOp(
                            set_value(5),
                            Mult(),
                            set_value(5),
                        ),
                        "doc": ["5", "b"],
                    },
                ),
                infer_type=True,
                word_wrap=True,
            ),
            ("adder", {"default": "```(5 * 5)```", "doc": "5b"}),
        )

    def test_from_docstring_numpydoc_only_returns(self) -> None:
        """
        Tests whether `docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_numpydoc_only_returns_str`"""
        ir, returns = parse.docstring(
            docstring_numpydoc_only_returns_str,
            return_tuple=True,
            emit_default_doc=False,
        )
        self.assertTrue(returns)
        self.assertDictEqual(
            ir,
            {
                "doc": ir["doc"],
                "name": None,
                "params": OrderedDict(),
                "returns": intermediate_repr_no_default_doc["returns"],
                "type": "static",
            },
        )

    def test_from_docstring_numpydoc_only_doc_str(self) -> None:
        """
        Tests whether `docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_numpydoc_only_doc_str`"""
        ir, returns = parse.docstring(
            docstring_numpydoc_only_doc_str.strip(), return_tuple=True
        )
        self.assertFalse(returns)
        self.assertDictEqual(
            ir,
            {
                "doc": intermediate_repr_no_default_doc["doc"],
                "name": None,
                "params": OrderedDict(),
                "returns": None,
                "type": "static",
            },
        )

    def test_from_docstring_google_str(self) -> None:
        """
        Tests whether `parse_docstring` produces `intermediate_repr_no_default_doc`
              from `docstring_google_str`
        """
        ir = parse_docstring(docstring_google_str)
        _intermediate_repr_no_default_doc = deepcopy(intermediate_repr_no_default_doc)
        _intermediate_repr_no_default_doc["doc"] = docstring_header_str
        self.assertDictEqual(ir, _intermediate_repr_no_default_doc)

    def test_from_docstring_google_tf_squared_hinge(self) -> None:
        """
        Tests whether `parse_docstring` produces the right IR
              from `docstring_google_tf_squared_hinge_str`
        """
        self.assertDictEqual(
            parse_docstring(
                docstring_google_tf_squared_hinge_str,
                emit_default_doc=True,
                infer_type=True,
                default_search_announce=("Default value is", "defaults to"),
            ),
            docstring_google_tf_squared_hinge_ir,
        )

    def test_from_docstring_google_tf_adam(self) -> None:
        """
        Tests whether `parse_docstring` produces the right IR
              from `docstring_google_tf_squared_hinge_str`
        """
        self.assertDictEqual(
            parse_docstring(
                docstring_google_tf_adam_str, emit_default_doc=True, infer_type=True
            ),
            docstring_google_tf_adam_ir,
        )

    def test_from_docstring_google_tf_adadelta_str(self) -> None:
        """
        Tests whether `parse_docstring` produces the right IR
              from `docstring_google_tf_adadelta_str`
        """
        self.assertDictEqual(
            parse_docstring(
                docstring_google_tf_adadelta_str, emit_default_doc=True, infer_type=True
            ),
            docstring_google_tf_adadelta_ir,
        )

    def test_from_docstring_google_tf_lambda_callback_str(self) -> None:
        """
        Tests whether `parse_docstring` produces the right IR
              from `docstring_google_tf_lambda_callback_str`
        """
        self.assertDictEqual(
            parse_docstring(
                docstring_google_tf_lambda_callback_str,
                emit_default_doc=True,
                infer_type=True,
                parse_original_whitespace=True,
            ),
            docstring_google_tf_lambda_callback_ir,
        )

    def test_from_docstring_google_pytorch_lbfgs_str(self) -> None:
        """
        Tests whether `parse_docstring` produces the right IR
              from `docstring_google_pytorch_lbfgs_str`
        """
        self.assertDictEqual(
            parse_docstring(
                docstring_google_pytorch_lbfgs_str,
                emit_default_doc=False,
                infer_type=True,
                parse_original_whitespace=False,
            ),
            docstring_google_pytorch_lbfgs_ir,
        )

    def test_docstring_header_and_return_str(self) -> None:
        """Tests that `docstring_header_and_return_str` can produce IR"""
        _intermediate_repr = deepcopy(intermediate_repr_no_default_doc)
        _intermediate_repr["params"] = OrderedDict()
        self.assertDictEqual(
            parse_docstring(docstring_header_and_return_str), _intermediate_repr
        )


unittest_main()



""" Tests for CLI doctrans subparser (__main__.py) """
from os import path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, run_cli_test, unittest_main


class TestCliDocTrans(TestCase):
    """Test class for __main__.py"""

    def test_doctrans_fails_with_wrong_args(self) -> None:
        """Tests CLI interface wrong args failure case"""

        run_cli_test(
            self,
            ["doctrans", "--wrong"],
            exit_code=2,
            output="the following arguments are required: --filename, --format\n",
        )

    def test_doctrans_fails_with_file_missing(self) -> None:
        """Tests CLI interface file missing failure case"""

        with patch("cdd.__main__.doctrans", mock_function):
            self.assertTrue(
                run_cli_test(
                    self,
                    [
                        "doctrans",
                        "--filename",
                        "foo",
                        "--format",
                        "google",
                        "--no-type-annotations",
                    ],
                    exit_code=2,
                    output="--filename must be an existent file. Got: 'foo'\n",
                ),
            )

    def test_doctrans_succeeds(self) -> None:
        """Tests CLI interface gets all the way to the doctrans call without error"""

        with TemporaryDirectory() as tempdir:
            filename = path.join(tempdir, "foo")
            open(filename, "a").close()
            with patch("cdd.__main__.doctrans", mock_function):
                self.assertTrue(
                    run_cli_test(
                        self,
                        [
                            "doctrans",
                            "--filename",
                            filename,
                            "--format",
                            "numpydoc",
                            "--type-annotations",
                        ],
                        exit_code=None,
                        output=None,
                    ),
                )


unittest_main()



"""
Mocks for docstrings
"""
from textwrap import indent

from cdd.pure_utils import identity, tab

docstring_header_str = (
    "Acquire from the official tensorflow_datasets model zoo,"
    " or the ophthalmology focussed ml-prepare\n"
)
# "library".strip()

docstring_header_and_return_str = "\n".join(
    (
        docstring_header_str,
        ":returns: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))",
        ":rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```",
    )
)

docstring_extra_colons_str = """
Some comment

:param dataset_name: Example: foo
:type dataset_name: ```str```
"""

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/losses.py#L845-L858
docstring_google_tf_squared_hinge_no_args_doc_str = (
    "Initializes `SquaredHinge` instance."
)
docstring_google_tf_squared_hinge = (
    docstring_google_tf_squared_hinge_no_args_doc_str,
    "Args:",
    "  reduction: (Optional) Type of `tf.keras.losses.Reduction` to apply to",
    "    loss. Default value is `AUTO`. `AUTO` indicates that the reduction",
    "    option will be determined by the usage context. For almost all cases",
    "    this defaults to `SUM_OVER_BATCH_SIZE`. When used with",
    "    `tf.distribute.Strategy`, outside of built-in training loops such as",
    "    `tf.keras` `compile` and `fit`, using `AUTO` or `SUM_OVER_BATCH_SIZE`",
    "    will raise an error. Please see this custom training [tutorial](",
    "      https://www.tensorflow.org/tutorials/distribute/custom_training)",
    "    for more details.",
    "  name: Optional name for the op. Defaults to 'squared_hinge'.",
)
docstring_google_tf_squared_hinge_str = "\n".join(docstring_google_tf_squared_hinge)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/optimizer_v2/adam.py#L35-L103
docstring_google_tf_adam = (
    "Optimizer that implements the Adam algorithm.",
    "Adam optimization is a stochastic gradient descent method that is based on",
    "adaptive estimation of first-order and second-order moments.",
    "According to",
    "[Kingma et al., 2014](http://arxiv.org/abs/1412.6980),",
    'the method is "*computationally',
    "efficient, has little memory requirement, invariant to diagonal rescaling of",
    "gradients, and is well suited for problems that are large in terms of",
    'data/parameters*".',
    "Args:",
    "learning_rate: A `Tensor`, floating point value, or a schedule that is a",
    "  `tf.keras.optimizers.schedules.LearningRateSchedule`, or a callable",
    "  that takes no arguments and returns the actual value to use, The",
    "  learning rate. Defaults to 0.001.",
    "beta_1: A float value or a constant float tensor, or a callable",
    "  that takes no arguments and returns the actual value to use. The",
    "  exponential decay rate for the 1st moment estimates. Defaults to 0.9.",
    "beta_2: A float value or a constant float tensor, or a callable",
    "  that takes no arguments and returns the actual value to use, The",
    "  exponential decay rate for the 2nd moment estimates. Defaults to 0.999.",
    "epsilon: A small constant for numerical stability. This epsilon is",
    '  "epsilon hat" in the Kingma and Ba paper (in the formula just before',
    "  Section 2.1), not the epsilon in Algorithm 1 of the paper. Defaults to",
    "  1e-7.",
    "amsgrad: Boolean. Whether to apply AMSGrad variant of this algorithm from",
    '  the paper "On the Convergence of Adam and beyond". Defaults to `False`.',
    "name: Optional name for the operations created when applying gradients.",
    '  Defaults to `"Adam"`.',
    "**kwargs: Keyword arguments. Allowed to be one of",
    '  `"clipnorm"` or `"clipvalue"`.',
    '  `"clipnorm"` (float) clips gradients by norm; `"clipvalue"` (float) clips',
    "  gradients by value.",
    "  ",
    "Usage:",
    ">>> opt = tf.keras.optimizers.Adam(learning_rate=0.1)",
    ">>> var1 = tf.Variable(10.0)",
    ">>> loss = lambda: (var1 ** 2)/2.0       # d(loss)/d(var1) == var1",
    ">>> step_count = opt.minimize(loss, [var1]).numpy()",
    ">>> # The first step is `-learning_rate*sign(grad)`",
    ">>> var1.numpy()",
    "9.9",
    "Reference:",
    "- [Kingma et al., 2014](http://arxiv.org/abs/1412.6980)",
    "- [Reddi et al., 2018](",
    "    https://openreview.net/pdf?id=ryQu7f-RZ) for `amsgrad`.",
    "Notes:",
    "The default value of 1e-7 for epsilon might not be a good default in",
    "general. For example, when training an Inception network on ImageNet a",
    "current good choice is 1.0 or 0.1. Note that since Adam uses the",
    "formulation just before Section 2.1 of the Kingma and Ba paper rather than",
    'the formulation in Algorithm 1, the "epsilon" referred to here is "epsilon',
    'hat" in the paper.',
    "The sparse implementation of this algorithm (used when the gradient is an",
    "IndexedSlices object, typically because of `tf.gather` or an embedding",
    "lookup in the forward pass) does apply momentum to variable slices even if",
    "they were not used in the forward pass (meaning they have a gradient equal",
    "to zero). Momentum decay (beta1) is also applied to the entire momentum",
    "accumulator. This means that the sparse behavior is equivalent to the dense",
    "behavior (in contrast to some momentum implementations which ignore momentum",
    "unless a variable slice was actually used).",
)
docstring_google_tf_adam_str = "\n".join(docstring_google_tf_adam)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/optimizer_v2/adadelta.py#L33-L77
docstring_google_tf_adadelta = (
    "Optimizer that implements the Adadelta algorithm.",
    "",
    "Adadelta optimization is a stochastic gradient descent method that is based on",
    "adaptive learning rate per dimension to address two drawbacks:",
    "",
    "- The continual decay of learning rates throughout training",
    "- The need for a manually selected global learning rate",
    "",
    "Adadelta is a more robust extension of Adagrad that adapts learning rates",
    "based on a moving window of gradient updates, instead of accumulating all",
    "past gradients. This way, Adadelta continues learning even when many updates",
    "have been done. Compared to Adagrad, in the original version of Adadelta you",
    "don't have to set an initial learning rate. In this version, initial",
    "learning rate can be set, as in most other Keras optimizers.",
    "",
    'According to section 4.3 ("Effective Learning rates"), near the end of',
    "training step sizes converge to 1 which is effectively a high learning",
    "rate which would cause divergence. This occurs only near the end of the",
    "training as gradients and step sizes are small, and the epsilon constant",
    "in the numerator and denominator dominate past gradients and parameter",
    "updates which converge the learning rate to 1.",
    "",
    'According to section 4.4("Speech Data"),where a large neural network with',
    "4 hidden layers was trained on a corpus of US English data, ADADELTA was",
    "used with 100 network replicas.The epsilon used is 1e-6 with rho=0.95",
    "which converged faster than ADAGRAD, by the following construction:",
    "def __init__(self, lr=1.0, rho=0.95, epsilon=1e-6, decay=0., **kwargs):",
    "",
    "Args:",
    "learning_rate: A `Tensor`, floating point value, or a schedule that is a",
    "  `tf.keras.optimizers.schedules.LearningRateSchedule`. The learning rate.",
    "  To match the exact form in the original paper use 1.0.",
    "rho: A `Tensor` or a floating point value. The decay rate.",
    "epsilon: A `Tensor` or a floating point value.  A constant epsilon used",
    "         to better conditioning the grad update.",
    "name: Optional name prefix for the operations created when applying",
    '  gradients.  Defaults to `"Adadelta"`.',
    "**kwargs: Keyword arguments. Allowed to be one of",
    '  `"clipnorm"` or `"clipvalue"`.',
    '  `"clipnorm"` (float) clips gradients by norm; `"clipvalue"` (float) clips',
    "  gradients by value.",
    "",
    "Reference:",
    "- [Zeiler, 2012](http://arxiv.org/abs/1212.5701)",
)
docstring_google_tf_adadelta_str = "\n".join(docstring_google_tf_adadelta)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/callbacks.py#L2649-L2699
docstring_google_tf_lambda_callback = (
    "Callback for creating simple, custom callbacks on-the-fly.",
    "",
    "  This callback is constructed with anonymous functions that will be called",
    "  at the appropriate time. Note that the callbacks expects positional",
    "  arguments, as:",
    "",
    "  - `on_epoch_begin` and `on_epoch_end` expect two positional arguments:",
    "    `epoch`, `logs`",
    "  - `on_batch_begin` and `on_batch_end` expect two positional arguments:",
    "    `batch`, `logs`",
    "  - `on_train_begin` and `on_train_end` expect one positional argument:",
    "    `logs`",
    "",
    "  Args:",
    "      on_epoch_begin: called at the beginning of every epoch.",
    "      on_epoch_end: called at the end of every epoch.",
    "      on_batch_begin: called at the beginning of every batch.",
    "      on_batch_end: called at the end of every batch.",
    "      on_train_begin: called at the beginning of model training.",
    "      on_train_end: called at the end of model training.",
    "",
    "  Example:",
    "",
    "  ```python",
    "  # Print the batch number at the beginning of every batch.",
    "  batch_print_callback = LambdaCallback(",
    "      on_batch_begin=lambda batch,logs: print(batch))",
    "",
    "  # Stream the epoch loss to a file in JSON format. The file content",
    "  # is not well-formed JSON but rather has a JSON object per line.",
    "  import json",
    "  json_log = open('loss_log.json', mode='wt', buffering=1)",
    "  json_logging_callback = LambdaCallback(",
    "      on_epoch_end=lambda epoch, logs: json_log.write(",
    "          json.dumps({'epoch': epoch, 'loss': logs['loss']}) + '\\n'),",
    "      on_train_end=lambda logs: json_log.close()",
    "  )",
    "",
    "  # Terminate some processes after having finished model training.",
    "  processes = ...",
    "  cleanup_callback = LambdaCallback(",
    "      on_train_end=lambda logs: [",
    "          p.terminate() for p in processes if p.is_alive()])",
    "",
    "  model.fit(...,",
    "            callbacks=[batch_print_callback,",
    "                       json_logging_callback,",
    "                       cleanup_callback])",
    "  ```",
)
docstring_google_tf_lambda_callback_str = "\n".join(docstring_google_tf_lambda_callback)

docstring_google_pytorch_lbfgs = (
    "",
    "Implements L-BFGS algorithm, heavily inspired by `minFunc",
    "    <https://www.cs.ubc.ca/~schmidtm/Software/minFunc.html>`.",
    "",
    "    Args:",
    "        lr (float): learning rate (default: 1)",
    "        max_iter (int): maximal number of iterations per optimization step",
    "            (default: 20)",
    "        max_eval (int): maximal number of function evaluations per optimization",
    "            step (default: max_iter * 1.25).",
    "        tolerance_grad (float): termination tolerance on first order optimality",
    "            (default: 1e-5).",
    "        tolerance_change (float): termination tolerance on function",
    "            value/parameter changes (default: 1e-9).",
    "        history_size (int): update history size (default: 100).",
    "        line_search_fn (str): either 'strong_wolfe' or None (default: None).",
)
docstring_google_pytorch_lbfgs_str = "\n".join(docstring_google_pytorch_lbfgs)

docstring_google_str = """{header_doc_str}

Args:
  dataset_name (str): name of dataset. Defaults to "mnist"
  tfds_dir (str): directory to look for models in. Defaults to "~/tensorflow_datasets"
  K (Literal['np', 'tf']): backend engine, e.g., `np` or `tf`. Defaults to "np"
  as_numpy (Optional[bool]): Convert to numpy ndarrays. Defaults to None
  data_loader_kwargs (Optional[dict]): pass this as arguments to data_loader function

Returns:
  Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]:
   Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
""".format(
    header_doc_str=docstring_header_str
)

docstring_no_default_doc_str = """
{header_doc_str}
:param dataset_name: name of dataset.
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in.
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`.
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays.
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:returns: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
""".format(
    header_doc_str=docstring_header_str
)

docstring_no_default_doc_wrapped_str = docstring_no_default_doc_str.replace(
    " np.ndarray]]```", "\n{tab}np.ndarray]]```".format(tab=tab)
)

docstring_no_default_str = """
{header_doc_str}
:param dataset_name: name of dataset.
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in.
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`.
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays.
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:returns: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
""".format(
    header_doc_str=docstring_header_str
)

docstring_numpydoc_only_doc_str = """
{header_doc_str}
""".format(
    header_doc_str=docstring_header_str
)

docstring_numpydoc_only_params_str = """
Parameters
----------
dataset_name : str
    name of dataset. Defaults to "mnist"
tfds_dir : str
    directory to look for models in. Defaults to "~/tensorflow_datasets"
K : Literal['np', 'tf']
    backend engine, e.g., `np` or `tf`. Defaults to "np"
as_numpy : Optional[bool]
    Convert to numpy ndarrays.
data_loader_kwargs : Optional[dict]
    pass this as arguments to data_loader function
"""

docstring_numpydoc_only_returns_str = """
Returns
-------
Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]
    Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
"""

docstring_numpydoc_str = """
{header_doc_str}

Parameters
----------
dataset_name : str
    name of dataset. Defaults to "mnist"
tfds_dir : str
    directory to look for models in. Defaults to "~/tensorflow_datasets"
K : Literal['np', 'tf']
    backend engine, e.g., `np` or `tf`. Defaults to "np"
as_numpy : Optional[bool]
    Convert to numpy ndarrays. Defaults to None
data_loader_kwargs : Optional[dict]
    pass this as arguments to data_loader function

{docstring_numpydoc_only_returns_str}""".format(
    header_doc_str=docstring_header_str,
    docstring_numpydoc_only_returns_str=docstring_numpydoc_only_returns_str.lstrip(
        "\n"
    ),
)

docstring_only_return_type_str = """
Some comment

:param dataset_name: Example: foo

:returns: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
"""

docstring_str = """
{header_doc_str}
:param dataset_name: name of dataset. Defaults to "mnist"
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`. Defaults to "np"
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays. Defaults to None
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:returns: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
""".format(
    header_doc_str=docstring_header_str
)

docstring_wrapped_str = docstring_str.replace(
    " np.ndarray]]```", "\n{tab}np.ndarray]]```".format(tab=tab)
)

docstring_no_type_str = """
{header_doc_str}

:param dataset_name: name of dataset. Defaults to "mnist"

:param tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"

:param K: backend engine, e.g., `np` or `tf`. Defaults to "np"

:param as_numpy: Convert to numpy ndarrays. Defaults to None

:param data_loader_kwargs: pass this as arguments to data_loader function

:returns: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
""".format(
    header_doc_str=docstring_header_str
)

docstring_no_type_no_default_str = """
{header_doc_str}
:param dataset_name: name of dataset.

:param tfds_dir: directory to look for models in.

:param K: backend engine, e.g., `np` or `tf`.

:param as_numpy: Convert to numpy ndarrays.

:param data_loader_kwargs: pass this as arguments to data_loader function

:returns: Train and tests dataset splits.
""".format(
    header_doc_str=docstring_header_str
)

docstring_repr_str = (
    indent(
        "\n".join(
            (
                "",
                "Emit a string representation of the current instance",
                "",
                ":returns: String representation of instance",
                ":rtype: ```str```",
                "",
            )
        ),
        tab * 2,
        identity,
    )
    + tab * 2
)

# docstring_repr_google_str = emit.docstring(parse.docstring(docstring_repr_str), docstring_format="google")
docstring_repr_google_str = (
    "\nEmit a string representation of the current instance\n\n\n\n\n"
    "Returns:\n"
    "  str:\n"
    "   String representation of instance\n"
)

__all__ = [
    "docstring_extra_colons_str",
    "docstring_google_str",
    "docstring_google_tf_adadelta_str",
    "docstring_google_tf_adam_str",
    "docstring_google_tf_lambda_callback_str",
    "docstring_google_tf_squared_hinge_str",
    "docstring_header_str",
    "docstring_no_default_doc_str",
    "docstring_numpydoc_only_doc_str",
    "docstring_numpydoc_only_params_str",
    "docstring_numpydoc_only_returns_str",
    "docstring_numpydoc_str",
    "docstring_only_return_type_str",
    "docstring_repr_str",
    "docstring_str",
]



"""
Tests route parsing
"""
from unittest import TestCase

from cdd.routes import parse
from cdd.routes.emit import create_util
from cdd.routes.emit_constants import create_helper_variants
from cdd.tests.mocks.openapi import openapi_dict
from cdd.tests.mocks.routes import (
    create_route,
    destroy_route,
    read_route,
    route_config,
    route_mock_prelude,
)
from cdd.tests.utils_for_tests import inspectable_compile, unittest_main


class TestRouteEmit(TestCase):
    """Tests `routes.parse`"""

    route_id_url = "{route_config[route]}/{{{route_config[primary_key]}}}".format(
        route_config=route_config
    )

    def test_create(self) -> None:
        """
        Tests whether `create_route` is produced by `emit.route`
        """
        _create_route = inspectable_compile(route_mock_prelude + create_route).create
        self.assertDictEqual(
            parse.bottle(_create_route),
            openapi_dict["paths"][route_config["route"]]["post"],
        )

    def test_create_util(self) -> None:
        """
        Tests whether `create_util` is produced by `create_util`
        """
        self.assertEqual(
            create_util(name=route_config["name"], route=route_config["route"]),
            create_helper_variants[-1].format(
                name=route_config["name"], route=route_config["route"]
            ),
        )

    def test_read(self) -> None:
        """
        Tests whether `read_route` is produced by `emit.route`
        """
        _read_route = inspectable_compile(route_mock_prelude + read_route).read
        self.assertDictEqual(
            parse.bottle(_read_route), openapi_dict["paths"][self.route_id_url]["get"]
        )

    def test_delete(self) -> None:
        """
        Tests whether `destroy_route` is produced by `emit.route`
        """
        _destroy_route = inspectable_compile(route_mock_prelude + destroy_route).destroy
        self.assertDictEqual(
            parse.bottle(_destroy_route),
            openapi_dict["paths"][self.route_id_url]["delete"],
        )


unittest_main()



"""
Mocks for methods
"""
from ast import (
    Attribute,
    BinOp,
    Call,
    ClassDef,
    Expr,
    FunctionDef,
    If,
    Index,
    Load,
    Mult,
    Name,
    Pass,
    Return,
    Subscript,
    Tuple,
    arguments,
    fix_missing_locations,
)
from functools import partial
from operator import add
from textwrap import indent

from cdd.ast_utils import maybe_type_comment, set_arg, set_slice, set_value
from cdd.pure_utils import emit_separating_tabs, tab
from cdd.tests.mocks.docstrings import (
    docstring_google_tf_adadelta,
    docstring_header_str,
    docstring_no_default_doc_wrapped_str,
    docstring_no_type_no_default_str,
    docstring_str,
)

return_ast = Return(
    value=Tuple(
        ctx=Load(),
        elts=[
            Call(
                args=[set_value(0)],
                func=Attribute(
                    Name("np", Load()),
                    "empty",
                    Load(),
                ),
                keywords=[],
                expr=None,
                expr_func=None,
            )
        ]
        * 2,
        expr=None,
    ),
    expr=None,
)

class_with_method_str = '''
class C(object):
    """ C class (mocked!) """

    def function_name(
        self,
        dataset_name="mnist",
        tfds_dir="~/tensorflow_datasets",
        K="np",
        as_numpy=None,
        **data_loader_kwargs
    ):
        """
{header_doc_str}{sep}
        :param dataset_name: name of dataset.
        :type dataset_name: ```str```
{sep}
        :param tfds_dir: directory to look for models in.
        :type tfds_dir: ```str```
{sep}
        :param K: backend engine, e.g., `np` or `tf`.
        :type K: ```Literal['np', 'tf']```
{sep}
        :param as_numpy: Convert to numpy ndarrays.
        :type as_numpy: ```Optional[bool]```
{sep}
        :param data_loader_kwargs: pass this as arguments to data_loader function
        :type data_loader_kwargs: ```Optional[dict]```
{sep}
        :returns: Train and tests dataset splits.
        :rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray,
{indent}np.ndarray]]```
        """
        return np.empty(0), np.empty(0)
'''.format(
    header_doc_str=indent(docstring_header_str, tab * 2),
    sep=tab * 2,
    indent=" " * 12,
)

class_with_method_types_str = '''
class C(object):
    """ C class (mocked!) """

    def function_name(
        self,
        dataset_name: str = "mnist",
        tfds_dir: str = "~/tensorflow_datasets",
        K: Literal["np", "tf"] = "np",
        as_numpy: Optional[bool] = None,
        **data_loader_kwargs
    ) -> Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]:
        """
{header_doc_str}
        :param dataset_name: name of dataset.

        :param tfds_dir: directory to look for models in.

        :param K: backend engine, e.g., `np` or `tf`.

        :param as_numpy: Convert to numpy ndarrays.

        :param data_loader_kwargs: pass this as arguments to data_loader function

        :returns: Train and tests dataset splits.
        """
        return np.empty(0), np.empty(0)
'''.format(
    header_doc_str=indent(docstring_header_str, tab * 2)
)

class_with_method_and_body_types_str = '''
class C(object):
    """ C class (mocked!) """

    def function_name(
        self,
        dataset_name: str = "mnist",
        tfds_dir: Optional[str] = "~/tensorflow_datasets",
        K: Literal["np", "tf"] = "np",
        as_numpy: Optional[bool] = None,
        **data_loader_kwargs
    ) -> Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]:
        """
        {header_doc_str}

        :param dataset_name: name of dataset.

        :param tfds_dir: directory to look for models in.

        :param K: backend engine, e.g., `np` or `tf`.

        :param as_numpy: Convert to numpy ndarrays.

        :param data_loader_kwargs: pass this as arguments to data_loader function

        :returns: Train and tests dataset splits.
        """
        # some comment
        print(5 * 5)
        if True:
            print(True)
            return 5
        return np.empty(0), np.empty(0)
'''.format(
    header_doc_str=indent(docstring_header_str, tab * 2)
)

class_with_optional_arg_method_str = '''
class C(object):
    """ C class (mocked!) """

    def function_name(
        self,
        dataset_name: str,
        K: Optional[Literal["np", "tf"]] = None
    ) -> Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]:
        """
        {header_doc_str}

        :param dataset_name: name of dataset.

        :param K: backend engine, e.g., `np` or `tf`.

        :returns: Train and tests dataset splits.
        """
        return np.empty(0), np.empty(0)
'''.format(
    header_doc_str=indent(docstring_header_str, tab * 2)
)


returns_subscript = Subscript(
    Name("Union", Load()),
    set_slice(
        Tuple(
            [
                Subscript(
                    Name("Tuple", Load()),
                    set_slice(
                        Tuple(
                            [
                                Attribute(
                                    Attribute(
                                        Name("tf", Load()),
                                        "data",
                                        Load(),
                                    ),
                                    "Dataset",
                                    Load(),
                                )
                            ]
                            * 2,
                            Load(),
                            expr=None,
                        )
                    ),
                    Load(),
                ),
                Subscript(
                    Name("Tuple", Load()),
                    set_slice(
                        Tuple(
                            [
                                Attribute(
                                    Name("np", Load()),
                                    "ndarray",
                                    Load(),
                                )
                            ]
                            * 2,
                            Load(),
                            expr=None,
                        )
                    ),
                    Load(),
                ),
            ],
            Load(),
        )
    ),
    Load(),
)


class_with_method_and_body_types_ast = fix_missing_locations(
    ClassDef(
        name="C",
        bases=[Name("object", Load())],
        keywords=[],
        body=[
            Expr(set_value(" C class (mocked!) ")),
            FunctionDef(
                name="function_name",
                args=arguments(
                    posonlyargs=[],
                    vararg=None,
                    args=[
                        set_arg("self"),
                        set_arg(arg="dataset_name", annotation=Name("str", Load())),
                        set_arg(
                            arg="tfds_dir",
                            annotation=Subscript(
                                Name("Optional", Load()),
                                set_slice(Name("str", Load())),
                                Load(),
                            ),
                        ),
                        set_arg(
                            arg="K",
                            annotation=Subscript(
                                Name("Literal", Load()),
                                set_slice(
                                    Tuple(
                                        elts=list(map(set_value, ("np", "tf"))),
                                        ctx=Load(),
                                        expr=None,
                                    )
                                ),
                                Load(),
                            ),
                        ),
                        set_arg(
                            arg="as_numpy",
                            annotation=Subscript(
                                Name("Optional", Load()),
                                set_slice(Name("bool", Load())),
                                Load(),
                            ),
                        ),
                    ],
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=set_arg("data_loader_kwargs"),
                    defaults=list(
                        map(set_value, ("mnist", "~/tensorflow_datasets", "np", None))
                    ),
                    arg=None,
                ),
                body=[
                    Expr(
                        set_value(
                            docstring_no_type_no_default_str,
                        )
                    ),
                    Expr(
                        Call(
                            func=Name("print", Load()),
                            args=[
                                BinOp(
                                    set_value(5),
                                    Mult(),
                                    set_value(5),
                                )
                            ],
                            keywords=[],
                            expr=None,
                            expr_func=None,
                        )
                    ),
                    If(
                        test=set_value(True),
                        body=[
                            Expr(
                                Call(
                                    func=Name("print", Load()),
                                    args=[set_value(True)],
                                    keywords=[],
                                    expr=None,
                                    expr_func=None,
                                )
                            ),
                            Return(
                                value=set_value(5),
                                expr=None,
                            ),
                        ],
                        orelse=[],
                        expr_test=None,
                        stmt=None,
                    ),
                    Return(
                        value=Tuple(
                            elts=[
                                Call(
                                    func=Attribute(Name("np", Load()), "empty", Load()),
                                    args=[set_value(0)],
                                    keywords=[],
                                    expr=None,
                                    expr_func=None,
                                )
                            ]
                            * 2,
                            ctx=Load(),
                            expr=None,
                        ),
                        expr=None,
                    ),
                ],
                decorator_list=[],
                returns=returns_subscript,
                arguments_args=None,
                identifier_name=None,
                stmt=None,
                lineno=None,
                **maybe_type_comment
            ),
        ],
        decorator_list=[],
        expr=None,
        identifier_name=None,
    )
)

class_with_method_ast = fix_missing_locations(
    ClassDef(
        bases=[Name("object", Load())],
        body=[
            Expr(
                set_value(
                    " C class (mocked!) ",
                )
            ),
            FunctionDef(
                args=arguments(
                    args=list(
                        map(
                            set_arg,
                            ("self", "dataset_name", "tfds_dir", "K", "as_numpy"),
                        )
                    ),
                    defaults=list(
                        map(set_value, ("mnist", "~/tensorflow_datasets", "np", None))
                    ),
                    kw_defaults=[],
                    kwarg=set_arg("data_loader_kwargs"),
                    kwonlyargs=[],
                    posonlyargs=[],
                    vararg=None,
                    arg=None,
                ),
                body=[
                    Expr(
                        set_value(
                            emit_separating_tabs(
                                indent(docstring_no_default_doc_wrapped_str, tab * 2), 2
                            )
                        )
                    ),
                    return_ast,
                ],
                decorator_list=[],
                name="function_name",
                returns=None,
                arguments_args=None,
                identifier_name=None,
                stmt=None,
                **maybe_type_comment
            ),
        ],
        decorator_list=[],
        keywords=[],
        name="C",
        expr=None,
        identifier_name=None,
    )
)

class_with_method_types_ast = fix_missing_locations(
    ClassDef(
        bases=[Name("object", Load())],
        body=[
            Expr(
                set_value(
                    " C class (mocked!) ",
                )
            ),
            FunctionDef(
                args=arguments(
                    args=[
                        set_arg("self"),
                        set_arg(
                            annotation=Name(
                                "str",
                                Load(),
                            ),
                            arg="dataset_name",
                        ),
                        set_arg(
                            annotation=Name(
                                "str",
                                Load(),
                            ),
                            arg="tfds_dir",
                        ),
                        set_arg(
                            annotation=Subscript(
                                Name(
                                    "Literal",
                                    Load(),
                                ),
                                Index(
                                    value=Tuple(
                                        ctx=Load(),
                                        elts=[
                                            set_value(
                                                "np",
                                            ),
                                            set_value(
                                                "tf",
                                            ),
                                        ],
                                        expr=None,
                                    )
                                ),
                                Load(),
                            ),
                            arg="K",
                        ),
                        set_arg(
                            annotation=Subscript(
                                Name(
                                    "Optional",
                                    Load(),
                                ),
                                Index(value=Name("bool", Load())),
                                Load(),
                            ),
                            arg="as_numpy",
                        ),
                    ],
                    defaults=list(
                        map(set_value, ("mnist", "~/tensorflow_datasets", "np", None))
                    ),
                    kw_defaults=[],
                    kwarg=set_arg("data_loader_kwargs"),
                    kwonlyargs=[],
                    posonlyargs=[],
                    vararg=None,
                    arg=None,
                ),
                body=[
                    Expr(set_value(indent(docstring_no_type_no_default_str, tab * 2))),
                    return_ast,
                ],
                decorator_list=[],
                name="function_name",
                returns=returns_subscript,
                arguments_args=None,
                identifier_name=None,
                stmt=None,
                **maybe_type_comment
            ),
        ],
        decorator_list=[],
        keywords=[],
        name="C",
        expr=None,
        identifier_name=None,
    )
)

class_with_optional_arg_method_ast = ClassDef(
    bases=[Name("object", Load())],
    body=[
        Expr(
            set_value(
                " C class (mocked!) ",
            )
        ),
        FunctionDef(
            args=arguments(
                args=[
                    set_arg("self"),
                    set_arg(
                        annotation=Name(
                            "str",
                            Load(),
                        ),
                        arg="dataset_name",
                    ),
                    set_arg(
                        annotation=Subscript(
                            Name(
                                "Optional",
                                Load(),
                            ),
                            Index(
                                value=Subscript(
                                    Name(
                                        "Literal",
                                        Load(),
                                    ),
                                    Index(
                                        value=Tuple(
                                            ctx=Load(),
                                            elts=list(
                                                map(
                                                    set_value,
                                                    (
                                                        "np",
                                                        "tf",
                                                    ),
                                                )
                                            ),
                                            expr=None,
                                        )
                                    ),
                                    Load(),
                                )
                            ),
                            Load(),
                        ),
                        arg="K",
                    ),
                ],
                defaults=[set_value(None)],
                kw_defaults=[],
                kwarg=None,
                kwonlyargs=[],
                posonlyargs=[],
                vararg=None,
                arg=None,
            ),
            body=[
                Expr(set_value(indent(docstring_str, tab * 2))),
                return_ast,
            ],
            decorator_list=[],
            name="function_name",
            returns=returns_subscript,
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            **maybe_type_comment
        ),
    ],
    decorator_list=[],
    keywords=[],
    name="C",
    expr=None,
    identifier_name=None,
)

function_adder_str = '''
def add_6_5(*, a=6, b=5):
    """
    :param a: first param
    :type a: ```int```

    :param b: second param
    :type b: ```int```

    :returns: Aggregated summation of `a` and `b`.
    :rtype: ```int```
    """
    return operator.add(a, b)
'''

function_adder_ast = FunctionDef(
    name="add_6_5",
    args=arguments(
        vararg=None,
        kwarg=None,
        posonlyargs=[],
        args=[],
        kwonlyargs=list(map(set_arg, ("a", "b"))),
        kw_defaults=list(map(set_value, (6, 5))),
        defaults=[],
        arg=None,
    ),
    body=[
        Expr(
            set_value(
                "\n    :param a: first param\n    "
                ":type a: ```int```\n\n    "
                ":param b: second param\n    "
                ":type b: ```int```\n\n    "
                ":returns: Aggregated summation of `a` and `b`.\n    "
                ":rtype: ```int```\n    ",
            )
        ),
        Return(
            value=Call(
                func=Attribute(Name("operator", Load()), "add", Load()),
                args=[Name("a", Load()), Name("b", Load())],
                keywords=[],
                expr=None,
                expr_func=None,
            ),
            expr=None,
        ),
    ],
    decorator_list=[],
    arguments_args=None,
    identifier_name=None,
    stmt=None,
)

function_default_complex_default_arg_str = (
    "def call_peril(dataset_name: str='mnist', writer=stdout):\n\tpass"
)

function_default_complex_default_arg_ast = FunctionDef(
    name="call_peril",
    args=arguments(
        args=[
            set_arg(
                annotation=Name(
                    "str",
                    Load(),
                ),
                arg="dataset_name",
            ),
            set_arg("writer"),
        ],
        defaults=[
            set_value("mnist"),
            Name("stdout", Load()),
        ],
        kw_defaults=[],
        kwarg=None,
        kwonlyargs=[],
        posonlyargs=[],
        vararg=None,
        arg=None,
    ),
    body=[Pass()],
    decorator_list=[],
    lineno=None,
    arguments_args=None,
    identifier_name=None,
    stmt=None,
)

method_complex_args_variety_str = '''
def call_cliff(
    self,
    dataset_name,
    *,
    as_numpy,
    K: Literal["np", "tf"],
    tfds_dir="~/tensorflow_datasets",
    writer=stdout,
    **kwargs
) -> Literal["np", "tf"]:
    """
    Call cliff

    :param dataset_name: name of dataset.

    :param as_numpy: Convert to numpy ndarrays.

    :param K: backend engine, e.g., `np` or `tf`.

    :param tfds_dir: directory to look for models in.

    :param writer: IO object to write out to

    :param **kwargs: additional keyword arguments

    :returns: backend engine
    """
    return K
'''

method_complex_args_variety_ast = FunctionDef(
    name="call_cliff",
    args=arguments(
        posonlyargs=[],
        args=list(map(set_arg, ("self", "dataset_name"))),
        kwonlyargs=[
            set_arg("as_numpy"),
            set_arg(
                arg="K",
                annotation=Subscript(
                    Name("Literal", Load()),
                    Tuple(
                        elts=list(
                            map(
                                set_value,
                                (
                                    "np",
                                    "tf",
                                ),
                            )
                        ),
                        ctx=Load(),
                        expr=None,
                    ),
                    Load(),
                ),
            ),
            set_arg("tfds_dir"),
            set_arg("writer"),
        ],
        kw_defaults=[
            None,
            None,
            set_value("~/tensorflow_datasets"),
            Name("stdout", Load()),
        ],
        kwarg=set_arg("kwargs"),
        defaults=[],
        arg=None,
    ),
    body=[
        Expr(
            set_value(
                "\n    Call cliff\n\n    "
                ":param dataset_name: name of dataset.\n\n    "
                ":param as_numpy: Convert to numpy ndarrays.\n\n    "
                ":param K: backend engine, e.g., `np` or `tf`.\n\n    "
                ":param tfds_dir: directory to look for models in.\n\n    "
                ":param writer: IO object to write out to\n\n    "
                ":param **kwargs: additional keyword arguments\n\n    "
                ":returns: backend engine\n    ",
            )
        ),
        Return(value=Name("K", Load()), expr=None),
    ],
    decorator_list=[],
    returns=Subscript(
        Name("Literal", Load()),
        Tuple(
            elts=list(
                map(
                    set_value,
                    (
                        "np",
                        "tf",
                    ),
                )
            ),
            ctx=Load(),
            expr=None,
        ),
        Load(),
    ),
    arguments_args=None,
    identifier_name=None,
    stmt=None,
)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/losses.py#L1327-L1355
function_google_tf_squared_hinge_docstring = (
    "Computes the squared hinge loss between `y_true` and `y_pred`.",
    "",
    "  `loss = mean(square(maximum(1 - y_true * y_pred, 0)), axis=-1)`",
    "",
    "  Standalone usage:",
    "",
    "  >>> y_true = np.random.choice([-1, 1], size=(2, 3))",
    "  >>> y_pred = np.random.random(size=(2, 3))",
    "  >>> loss = tf.keras.losses.squared_hinge(y_true, y_pred)",
    "  >>> assert loss.shape == (2,)",
    "  >>> assert np.array_equal(",
    "  ...     loss.numpy(),",
    "  ...     np.mean(np.square(np.maximum(1. - y_true * y_pred, 0.)), axis=-1))",
    "",
    "  Args:",
    "    y_true: The ground truth values. `y_true` values are expected to be -1 or 1.",
    "      If binary (0 or 1) labels are provided we will convert them to -1 or 1.",
    "      shape = `[batch_size, d0, .. dN]`.",
    "    y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.",
    "",
    "  Returns:",
    "     Squared hinge loss values. shape = `[batch_size, d0, .. dN-1]`.",
    "  ",
)
function_google_tf_squared_hinge_docstring_str = "\n".join(
    function_google_tf_squared_hinge_docstring
)
function_google_tf_squared_hinge = (
    "def squared_hinge(y_true, y_pred):",
    '  """{}"""'.format(function_google_tf_squared_hinge_docstring_str),
    "  y_pred = ops.convert_to_tensor_v2(y_pred)",
    "  y_true = math_ops.cast(y_true, y_pred.dtype)",
    "  y_true = _maybe_convert_labels(y_true)",
    "  return K.mean(",
    "      math_ops.square(math_ops.maximum(1. - y_true * y_pred, 0.)), axis=-1)",
)
function_google_tf_squared_hinge_str = "\n".join(function_google_tf_squared_hinge)

docstring_google_tf_adadelta_function = (
    "",
    "class Adadelta(object):",
    '  """{}"""'.format(
        "\n".join(map(partial(add, " " * 2), docstring_google_tf_adadelta))
    ),
    "",
    "  _HAS_AGGREGATE_GRAD = True",
    "",
    "  def __init__(self,",
    "               learning_rate=0.001,",
    "               rho=0.95,",
    "               epsilon=1e-7,",
    "               name='Adadelta',",
    "               **kwargs):",
    "    # super(Adadelta, self).__init__(name, **kwargs)",
    "    self._set_hyper('learning_rate', kwargs.get('lr', learning_rate))",
    "    self._set_hyper('decay', self._initial_decay)",
    "    self._set_hyper('rho', rho)",
    "    self.epsilon = epsilon or backend_config.epsilon()",
    "",
    "  def _create_slots(self, var_list):",
    "    # Separate for-loops to respect the ordering of slot variables from v1.",
    "    for v in var_list:",
    "      self.add_slot(v, 'accum_grad')",
    "    for v in var_list:",
    "      self.add_slot(v, 'accum_var')",
    "",
    "  def _prepare_local(self, var_device, var_dtype, apply_state):",
    "    # super(Adadelta, self)._prepare_local(var_device, var_dtype, apply_state)",
    "    apply_state[(var_device, var_dtype)].update(",
    "        dict(",
    "            epsilon=ops.convert_to_tensor_v2_with_dispatch(",
    "                self.epsilon, var_dtype),",
    "            rho=array_ops.identity(self._get_hyper('rho', var_dtype))))",
    "",
    "  def set_weights(self, weights):",
    "    params = self.weights",
    "    # Override set_weights for backward compatibility of Keras V1 optimizer",
    "    # since it does not include iteration at head of the weight list. Set",
    "    # iteration to 0.",
    "    if len(params) == len(weights) + 1:",
    "      weights = [np.array(0)] + weights",
    "    # super(Adadelta, self).set_weights(weights)",
    "",
    "  def _resource_apply_dense(self, grad, var, apply_state=None):",
    "    var_device, var_dtype = var.device, var.dtype.base_dtype",
    "    coefficients = ((apply_state or {}).get((var_device, var_dtype))",
    "                    or self._fallback_apply_state(var_device, var_dtype))",
    "",
    "    accum_grad = self.get_slot(var, 'accum_grad')",
    "    accum_var = self.get_slot(var, 'accum_var')",
    "    return gen_training_ops.ResourceApplyAdadelta(",
    "        var=var.handle,",
    "        accum=accum_grad.handle,",
    "        accum_update=accum_var.handle,",
    "        lr=coefficients['lr_t'],",
    "        rho=coefficients['rho'],",
    "        epsilon=coefficients['epsilon'],",
    "        grad=grad,",
    "        use_locking=self._use_locking)",
    "",
    "  def _resource_apply_sparse(self, grad, var, indices, apply_state=None):",
    "    var_device, var_dtype = var.device, var.dtype.base_dtype",
    "    coefficients = ((apply_state or {}).get((var_device, var_dtype))",
    "                    or self._fallback_apply_state(var_device, var_dtype))",
    "",
    "    accum_grad = self.get_slot(var, 'accum_grad')",
    "    accum_var = self.get_slot(var, 'accum_var')",
    "    return gen_training_ops.ResourceSparseApplyAdadelta(",
    "        var=var.handle,",
    "        accum=accum_grad.handle,",
    "        accum_update=accum_var.handle,",
    "        lr=coefficients['lr_t'],",
    "        rho=coefficients['rho'],",
    "        epsilon=coefficients['epsilon'],",
    "        grad=grad,",
    "        indices=indices,",
    "        use_locking=self._use_locking)",
    "",
    "  def get_config(self):",
    "    config = super(Adadelta, self).get_config()",
    "    config.update({",
    "        'learning_rate': self._serialize_hyperparameter('learning_rate'),",
    "        'decay': self._serialize_hyperparameter('decay'),",
    "        'rho': self._serialize_hyperparameter('rho'),",
    "        'epsilon': self.epsilon,",
    "    })",
    "    return config",
)
docstring_google_tf_adadelta_function_str = "\n".join(
    docstring_google_tf_adadelta_function
)

__all__ = [
    "class_with_method_and_body_types_ast",
    "class_with_method_and_body_types_str",
    "class_with_method_ast",
    "class_with_method_str",
    "class_with_method_types_ast",
    "class_with_method_types_str",
    "class_with_optional_arg_method_ast",
    "class_with_optional_arg_method_str",
    "docstring_google_tf_adadelta_function_str",
    "function_adder_ast",
    "function_adder_str",
    "function_default_complex_default_arg_ast",
    "function_default_complex_default_arg_str",
    "function_google_tf_squared_hinge_str",
    "method_complex_args_variety_ast",
    "method_complex_args_variety_str",
]



""" Tests for doctrans """
from os import path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cdd.doctrans import doctrans
from cdd.tests.mocks.methods import return_ast
from cdd.tests.utils_for_tests import unittest_main


class TestDocTrans(TestCase):
    """Test class for doctrans.py"""

    def test_doctrans(self) -> None:
        """Tests doctrans"""

        with TemporaryDirectory() as temp_dir, patch(
            "cdd.doctrans.DocTrans",
            lambda **kwargs: type(
                "DocTrans", tuple(), {"visit": lambda *args: return_ast}
            )(),
        ):
            filename = path.join(temp_dir, "foo")
            with open(filename, "wt") as f:
                f.write("5*5")
            self.assertIsNone(
                doctrans(
                    filename=filename,
                    docstring_format="numpydoc",
                    type_annotations=True,
                )
            )


unittest_main()



"""
Tests for docstring parsing
"""
from ast import Module
from collections import namedtuple
from io import StringIO
from typing import Any
from unittest import TestCase
from unittest.mock import MagicMock, patch

from cdd.pure_utils import PY_GTE_3_8
from cdd.tests.mocks.classes import tensorboard_doc_str, tensorboard_doc_str_no_args_str
from cdd.tests.utils_for_tests import remove_args_from_docstring, unittest_main


class TestUtilsForTests(TestCase):
    """
    Tests whether docstrings are parsed out—and emitted—correctly
    """

    def test_unittest_main(self) -> None:
        """
        Tests whether `unittest_main` is called when `__name__ == '__main__'`
        """
        self.assertEqual(type(unittest_main).__name__, "function")
        self.assertIsNone(unittest_main())
        argparse_mock = MagicMock()
        with patch("cdd.tests.utils_for_tests.__name__", "__main__"), patch(
            "sys.stderr", new_callable=StringIO
        ), self.assertRaises(SystemExit) as e:
            import cdd.tests.utils_for_tests

            cdd.tests.utils_for_tests.unittest_main()

        self.assertIsInstance(e.exception.code, bool)
        self.assertIsNone(argparse_mock.call_args)
        self.assertIsNone(cdd.tests.utils_for_tests.unittest_main())

    def test_run_ast_test(self) -> None:
        """
        Tests whether `run_ast_test` correct avoids running the AST comparison dependent on Python version
        """

        def assert_true(value, msg=None):
            """Version of `self.assertTrue` which also keeps count

            :param value: Potentially `True`
            :type value: ```Union[Literal[True], Any]```

            :param msg: Message to raise in error
            :type msg: ```Optional[str]```
            """
            TestUtilsForTests.i += TestUtilsForTests.increment
            assert value, msg or "{!r} not truthy".format(value)

        def assert_equal(a, b, msg=None):
            """Version of `self.assertEqual` which also keeps count

            :param a: Any value that can be compared. Compared with `b`.
            :type a: ```Any```

            :param b: Any value that can be compared. Compared with `a`.
            :type b: ```Any```

            :param msg: Message to raise in error
            :type msg: ```Optional[str]```
            """
            TestUtilsForTests.i += TestUtilsForTests.increment
            assert a == b, msg or "{!r} != {!r}".format(a, b)

        TestUtilsForTests.increment = 2 if PY_GTE_3_8 else 1
        TestUtilsForTests.i = 0

        test_case_module: Any = namedtuple("TestCase", ("assertTrue", "assertEqual"))(
            assert_true,
            assert_equal,
        )

        module = Module(body=[], type_ignores=[], stmt=None)

        import cdd.tests.utils_for_tests

        _orig_cdd_tests_utils_for_tests_PY3_8 = cdd.tests.utils_for_tests.PY3_8

        try:
            with patch("sys.version_info", (3, 7)):
                cdd.tests.utils_for_tests.PY_GTE_3_8 = (
                    cdd.tests.utils_for_tests.PY3_8
                ) = False

                cdd.tests.utils_for_tests.run_ast_test(test_case_module, module, module)
                self.assertEqual(TestUtilsForTests.increment * 2, TestUtilsForTests.i)

            with patch("sys.version_info", (3, 8)):
                cdd.tests.utils_for_tests.PY_GTE_3_8 = (
                    cdd.tests.utils_for_tests.PY3_8
                ) = True

                cdd.tests.utils_for_tests.run_ast_test(test_case_module, module, module)
                self.assertEqual(TestUtilsForTests.increment * 4, TestUtilsForTests.i)

            with patch("sys.version_info", (3, 9)):
                cdd.tests.utils_for_tests.PY_GTE_3_8 = (
                    cdd.tests.utils_for_tests.PY3_8
                ) = True

                cdd.tests.utils_for_tests.run_ast_test(test_case_module, module, module)
                self.assertEqual(TestUtilsForTests.increment * 6, TestUtilsForTests.i)
        finally:
            cdd.tests.utils_for_tests.PY3_8 = _orig_cdd_tests_utils_for_tests_PY3_8

    def test_remove_args_from_docstring(self) -> None:
        """
        Test that `remove_args_from_docstring`
         produces `tensorboard_doc_str_no_args_str`
             from `tensorboard_doc_str`
        """
        self.assertEqual(
            remove_args_from_docstring(tensorboard_doc_str),
            tensorboard_doc_str_no_args_str,
        )


unittest_main()
